import json
from typing import Any

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.output import (
    format_cmd_list,
    lines_to_str,
    smart_wrap_text,
)
from pcs.cli.common.parse_args import (
    FUTURE_OPTION,
    OUTPUT_FORMAT_VALUE_CMD,
    OUTPUT_FORMAT_VALUE_JSON,
    Argv,
    InputModifiers,
    KeyValueParser,
    ensure_unique_args,
    wait_to_timeout,
)
from pcs.cli.reports.output import (
    deprecation_warning,
    process_library_reports,
    warn,
)
from pcs.cli.resource.common import (
    check_is_not_stonith,
    get_resource_status_msg,
)
from pcs.cli.resource.output import (
    ResourcesConfigurationFacade,
    resources_to_cmd,
    resources_to_text,
)
from pcs.common import reports
from pcs.common.interface import dto
from pcs.common.pacemaker.resource.list import (
    CibResourcesDto,
    get_all_resources_ids,
    get_stonith_resources_ids,
)
from pcs.common.str_tools import format_list, format_plural
from pcs.lib.errors import LibraryError


def config(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    output = config_common(lib, argv, modifiers, stonith=False)
    if output:
        print(output)


def config_common(
    lib: Any, argv: Argv, modifiers: InputModifiers, stonith: bool
) -> str:
    """
    Also used by stonith commands.

    Options:
      * -f - CIB file
      * --output-format - supported formats: text, cmd, json
    """
    modifiers.ensure_only_supported("-f", output_format_supported=True)
    resources_facade = (
        ResourcesConfigurationFacade.from_resources_dto(
            lib.resource.get_configured_resources()
        )
        .filter_stonith(stonith)
        .filter_resources(argv)
    )
    output_format = modifiers.get_output_format()
    if output_format == OUTPUT_FORMAT_VALUE_CMD:
        output = format_cmd_list(
            [" \\\n".join(cmd) for cmd in resources_to_cmd(resources_facade)]
        )
    elif output_format == OUTPUT_FORMAT_VALUE_JSON:
        output = json.dumps(
            dto.to_dict(
                CibResourcesDto(
                    primitives=resources_facade.primitives,
                    clones=resources_facade.clones,
                    groups=resources_facade.groups,
                    bundles=resources_facade.bundles,
                )
            )
        )
    else:
        output = lines_to_str(
            smart_wrap_text(resources_to_text(resources_facade))
        )
    return output


def meta(lib: Any, argv: list[str], modifiers: InputModifiers) -> None:
    """
    Options:
      * --force - override editing connection attributes for guest nodes
      * --wait - wait for cluster to reach steady state
      * -f
    """
    modifiers.ensure_only_supported("-f", "--force", "--wait")
    modifiers.ensure_not_mutually_exclusive("-f", "--wait")
    wait_timeout = wait_to_timeout(modifiers.get("--wait"))
    force_flags = []
    if modifiers.get("--force"):
        force_flags.append(reports.codes.FORCE)
    if not argv:
        raise CmdLineInputError()
    resource_id = argv.pop(0)
    check_is_not_stonith(lib, [resource_id], "pcs stonith meta")
    meta_attrs_dict = KeyValueParser(argv).get_unique()

    lib.resource.update_meta(resource_id, meta_attrs_dict, force_flags)

    if wait_timeout >= 0:
        deprecation_warning(reports.messages.ResourceWaitDeprecated().message)
        lib.cluster.wait_for_pcmk_idle(wait_timeout)
        print(get_resource_status_msg(lib, resource_id))


def remove(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
      * --force - turn validation errors into warnings, (derecated) skip
                  resource stopping
      * --no-stop - don't stop resources before deletion
      * --future - specifying '--force' does not skip resource stopping
    """

    def _process_reports(
        report_list: reports.ReportItemList,
        force_flags: reports.types.ForceFlags,
    ) -> reports.ReportItemList:
        filtered_reports = []
        for report in report_list:
            if (
                report.message.code
                == reports.codes.CANNOT_REMOVE_RESOURCES_NOT_STOPPED
            ):
                continue
            if (
                report.severity.level == reports.ReportItemSeverity.ERROR
                and report.severity.force_code in force_flags
            ):
                report.severity = reports.ReportItemSeverity.warning()
            filtered_reports.append(report)
        return filtered_reports

    modifiers.ensure_only_supported("-f", "--force", FUTURE_OPTION, "--no-stop")
    modifiers.ensure_not_mutually_exclusive("-f", "--no-stop")

    if not argv:
        raise CmdLineInputError()
    ensure_unique_args(argv)

    resources_to_remove = set(argv)
    resources_dto = lib.resource.get_configured_resources()
    missing_ids = resources_to_remove - get_all_resources_ids(resources_dto)
    if missing_ids:
        raise CmdLineInputError(
            "Unable to find {resource}: {id_list}".format(
                resource=format_plural(missing_ids, "resource"),
                id_list=format_list(missing_ids),
            )
        )

    stonith_ids = resources_to_remove & get_stonith_resources_ids(resources_dto)
    if stonith_ids:
        raise CmdLineInputError(
            (
                "This command cannot remove stonith {resource}: {id_list}. Use "
                "'pcs stonith remove' instead."
            ).format(
                resource=format_plural(stonith_ids, "resource"),
                id_list=format_list(stonith_ids),
            )
        )

    force_flags = set()
    if modifiers.is_specified("--force"):
        force_flags.add(reports.codes.FORCE)

    if modifiers.is_specified("-f"):
        warn(
            "Resources are not going to be stopped before deletion because the "
            "command does not run on a live cluster"
        )
        lib.cib.remove_elements(resources_to_remove, force_flags)
        return

    dont_stop_me_now = modifiers.is_specified("--no-stop")
    if (
        not modifiers.is_specified(FUTURE_OPTION)
        and modifiers.is_specified("--force")
        and not dont_stop_me_now
    ):
        # deprecated after pcs-0.12.0
        deprecation_warning(
            "Using '--force' to skip resource stopping is deprecated and will "
            "be removed in a future release. Specify '--future' to switch to "
            "the future behavior and use '--no-stop' to skip resource stopping."
        )
        dont_stop_me_now = True

    if dont_stop_me_now:
        lib.cib.remove_elements(resources_to_remove, force_flags)
        return

    original_report_processor = lib.env.report_processor
    in_memory_report_processor = reports.processor.ReportProcessorInMemory()
    lib.env.report_processor = in_memory_report_processor

    try:
        # call without force flags the first time, so we really catch the
        # validation errors without deleting anything even when --force
        # was specified
        lib.cib.remove_elements(resources_to_remove)

        # the resources were successfully removed, but we still need to print
        # out the reports, since there could be INFO or DEBUG reports
        if in_memory_report_processor.reports:
            process_library_reports(
                in_memory_report_processor.reports,
                include_debug=modifiers.is_specified("--debug"),
            )
        return
    except LibraryError as e:
        filtered_reports = _process_reports(
            in_memory_report_processor.reports, force_flags
        )

        # if there are other errors than CANNOT_REMOVE_RESOURCES_NOT_STOPPED or
        # errors that would be forced, we can exit, since it does not make sense
        # to try stopping and removing the resources again
        if reports.has_errors(filtered_reports) or e.output or e.args:
            if filtered_reports:
                process_library_reports(
                    filtered_reports,
                    include_debug=modifiers.is_specified("--debug"),
                    exit_on_error=False,
                )
            raise e

    lib.env.report_processor = original_report_processor

    lib.resource.stop(resources_to_remove, force_flags)
    lib.cluster.wait_for_pcmk_idle(None)

    lib.cib.remove_elements(resources_to_remove, force_flags)



import json
import shlex
from unittest import TestCase

from pcs.common.interface import dto
from pcs.common.pacemaker.fencing_topology import (
    CibFencingLevelAttributeDto,
    CibFencingLevelNodeDto,
    CibFencingLevelRegexDto,
    CibFencingTopologyDto,
)

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner


class LevelsConfigJson(TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_empty(self):
        pcs_runner = PcsRunner(cib_file=get_test_resource("cib-empty.xml"))
        stdout, stderr, retval = pcs_runner.run(
            ["stonith", "level", "config", "--output-format=json"]
        )
        expected = CibFencingTopologyDto(
            target_node=[], target_regex=[], target_attribute=[]
        )
        self.assertEqual(
            json.loads(stdout), json.loads(json.dumps(dto.to_dict(expected)))
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_success(self):
        pcs_runner = PcsRunner(
            cib_file=get_test_resource("cib-fencing-levels.xml")
        )

        stdout, stderr, retval = pcs_runner.run(
            ["stonith", "level", "config", "--output-format=json"]
        )
        expected = CibFencingTopologyDto(
            [
                CibFencingLevelNodeDto(
                    id="fl-rh-1-1", target="rh-1", index=1, devices=["S1"]
                ),
                CibFencingLevelNodeDto(
                    id="fl-rh-1-2", target="rh-1", index=2, devices=["S2"]
                ),
                CibFencingLevelNodeDto(
                    id="fl-rh-2-1", target="rh-2", index=1, devices=["S3", "S4"]
                ),
            ],
            [
                CibFencingLevelRegexDto(
                    id="fl-rh-.-3",
                    target_pattern="rh-.*",
                    index=3,
                    devices=["S4"],
                )
            ],
            [
                CibFencingLevelAttributeDto(
                    id="fl-foo-4",
                    target_attribute="foo",
                    target_value="bar",
                    index=4,
                    devices=["S1", "S2", "S3"],
                )
            ],
        )
        self.assertEqual(
            json.loads(stdout), json.loads(json.dumps(dto.to_dict(expected)))
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)


class LevelsConfigCmd(TestCase):
    def setUp(self):
        self.old_cib = get_tmp_file("tier1_stonith_level_cmd_old")
        write_file_to_tmpfile(
            get_test_resource("cib-fencing-levels.xml"), self.old_cib
        )
        self.new_cib = get_tmp_file("tier1_stonith_level_cmd_new")
        write_file_to_tmpfile(get_test_resource("cib-empty.xml"), self.new_cib)
        self.old_pcs_runner = PcsRunner(self.old_cib.name)
        self.new_pcs_runner = PcsRunner(self.new_cib.name)
        self.maxDiff = None

    def tearDown(self):
        self.old_cib.close()
        self.new_cib.close()

    def _get_as_json(self, runner):
        stdout, stderr, retval = runner.run(
            ["stonith", "level", "config", "--output-format=json"]
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        return json.loads(stdout)

    def test_success(self):
        stdout, stderr, retval = self.old_pcs_runner.run(
            ["stonith", "level", "config", "--output-format=cmd"]
        )
        self.assertEqual(retval, 0)
        commands = [
            shlex.split(command)[1:]
            for command in stdout.replace("\\n", "").strip().split(";\n")
        ]

        for cmd in commands:
            stdout, stderr, retval = self.new_pcs_runner.run(cmd)
            self.assertEqual(
                retval,
                0,
                (
                    f"Command {cmd} exited with {retval}\nstdout:\n{stdout}\n"
                    f"stderr:\n{stderr}"
                ),
            )
        self.assertEqual(
            self._get_as_json(self.new_pcs_runner),
            self._get_as_json(self.old_pcs_runner),
        )


class LevelsConfigText(AssertPcsMixin, TestCase):
    FIXTURE_OUTPUT = (
        "Target (node): rh-1\n"
        "  Level 1: S1\n"
        "  Level 2: S2\n"
        "Target (node): rh-2\n"
        "  Level 1: S3 S4\n"
        "Target (regexp): rh-.*\n"
        "  Level 3: S4\n"
        "Target (attribute): foo=bar\n"
        "  Level 4: S1 S2 S3\n"
    )

    def setUp(self):
        self.maxDiff = None
        self.pcs_runner = PcsRunner(
            cib_file=get_test_resource("cib-fencing-levels.xml")
        )

    def test_empty(self):
        pcs_runner = PcsRunner(cib_file=get_test_resource("cib-empty.xml"))
        stdout, stderr, retval = pcs_runner.run(
            ["stonith", "level", "config", "--output-format=text"]
        )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_success(self):
        self.assert_pcs_success(
            ["stonith", "level", "config", "--output-format=text"],
            stdout_full=self.FIXTURE_OUTPUT,
        )

    def test_success_no_option(self):
        self.assert_pcs_success(
            ["stonith", "level", "config"], stdout_full=self.FIXTURE_OUTPUT
        )

    def test_success_no_config(self):
        self.assert_pcs_success(
            ["stonith", "level"], stdout_full=self.FIXTURE_OUTPUT
        )



from pprint import pformat
from urllib.parse import urlencode

from tornado.httputil import HTTPHeaders
from tornado.testing import AsyncHTTPTestCase
from tornado.web import Application

from pcs.daemon import ruby_pcsd

USER = "user"
GROUPS = ["group1", "group2"]
PASSWORD = "password"


class RubyPcsdWrapper(ruby_pcsd.Wrapper):
    def __init__(self, request_type):
        # pylint: disable=super-init-not-called
        self.request_type = request_type
        self.status_code = 200
        self.headers = {"Some": "value"}
        self.body = b"Success action"

    async def run_ruby(
        self,
        request_type,
        http_request=None,
        payload=None,
    ):
        del http_request, payload
        if request_type != self.request_type:
            raise AssertionError(
                f"Wrong request type: expected '{self.request_type}'"
                f" but was {request_type}"
            )
        return {
            "headers": self.headers,
            "status": self.status_code,
            "body": self.body,
        }


class AppTest(AsyncHTTPTestCase):
    wrapper = None

    def get_app(self):
        return Application(self.get_routes())

    def get_routes(self):
        # pylint: disable=no-self-use
        return []

    def fetch(self, path, raise_error=False, **kwargs):
        if "follow_redirects" not in kwargs:
            kwargs["follow_redirects"] = False

        if "is_ajax" in kwargs:
            if "headers" not in kwargs:
                kwargs["headers"] = {}
            kwargs["headers"]["X-Requested-With"] = "XMLHttpRequest"
            del kwargs["is_ajax"]

        response = super().fetch(path, raise_error=raise_error, **kwargs)
        # "Strict-Transport-Security" header is expected in every response
        self.assertTrue(
            "Strict-Transport-Security" in response.headers,
            f"No 'Strict-Transport-Security' header in response for '{path}'",
        )
        return response

    def post(self, path, body, **kwargs):
        kwargs.update(
            {
                "method": "POST",
                "body": urlencode(body),
            }
        )
        return self.fetch(path, **kwargs)

    def get(self, path, **kwargs):
        return self.fetch(path, **kwargs)

    def assert_headers_contains(self, headers: HTTPHeaders, contained: dict):
        self.assertTrue(
            all(item in headers.get_all() for item in contained.items()),
            "Headers does not contain expected headers"
            "\n  Expected headers:"
            f"\n    {pformat(contained, indent=6)}"
            "\n  All headers:"
            f"\n    {pformat(dict(headers.get_all()), indent=6)}",
        )

    def assert_wrappers_response(self, response):
        self.assertEqual(response.code, self.wrapper.status_code)
        self.assert_headers_contains(response.headers, self.wrapper.headers)
        self.assertEqual(response.body, self.wrapper.body)

    def assert_unauth_ajax(self, response):
        self.assertEqual(response.code, 401)
        self.assertEqual(response.body, b'{"notauthorized":"true"}')



from collections import defaultdict
from dataclasses import replace
from hashlib import sha1
from typing import Iterable, Optional, cast

from pcs.common.file_type_codes import FileTypeCode
from pcs.common.node_communicator import Communicator, RequestTarget
from pcs.common.reports import ReportItemContext
from pcs.common.reports.processor import ReportProcessor
from pcs.common.str_tools import format_list
from pcs.lib.communication.pcs_cfgsync import ConfigInfo, GetConfigs
from pcs.lib.communication.tools import run
from pcs.lib.file.instance import FileInstance
from pcs.lib.file.raw_file import RawFileError, raw_file_error_report
from pcs.lib.interface.config import (
    FacadeInterface,
    ParserErrorException,
    SyncVersionFacadeInterface,
)

from .const import SYNCED_CONFIGS


class ConfigFetcher:
    def __init__(
        self, node_communicator: Communicator, report_processor: ReportProcessor
    ):
        self._node_communicator = node_communicator
        self._report_processor = report_processor

    def fetch(
        self,
        cluster_name: str,
        target_list: Iterable[RequestTarget],
        file_type_codes_to_fetch: Iterable[FileTypeCode] = SYNCED_CONFIGS,
    ) -> tuple[dict[FileTypeCode, SyncVersionFacadeInterface], bool]:
        """
        Downloads configs from all nodes in the cluster, and find the newest
        among them. Returns a dict containing facade with the newest config for
        each received config type, and a bool indicating whether we were able to
        download configs from the cluster.

        cluster_name -- name of the cluster
        target_list -- list of request targets
        file_type_codes_to_fetch -- list of wanted file type codes
        """
        if not set(file_type_codes_to_fetch) <= set(SYNCED_CONFIGS):
            raise AssertionError(
                f"This method only supports {format_list(list(SYNCED_CONFIGS))}"
            )

        cmd = GetConfigs(self._report_processor, cluster_name, False)
        cmd.set_targets(target_list)  # type: ignore
        received_configs = run(self._node_communicator, cmd)  # type: ignore

        configs_to_update = {}

        for file_type in file_type_codes_to_fetch:
            if file_type not in received_configs.config_files:
                continue

            instance = FileInstance.for_common(file_type)
            configs = self._parse_received_configs(
                instance, received_configs.config_files[file_type]
            )
            newest_config = _find_newest_config(instance, configs)
            if newest_config is None:
                continue

            if not instance.raw_file.exists():
                # if the file does not exist locally, but we received it from
                # other nodes, then we want to save the file locally as well
                configs_to_update[file_type] = newest_config
                continue

            local_config = self._parse_local_config(instance)
            if local_config is None:
                # We were unable to parse the local config for this file_type
                # and we can't compare the received configs with it. But we
                # might still be able to properly work with the remaining
                # config types
                continue

            if local_config.data_version < newest_config.data_version:
                configs_to_update[file_type] = newest_config
            elif local_config.data_version > newest_config.data_version:
                pass
            elif _file_hash(instance, local_config) != _file_hash(
                instance, newest_config
            ):
                configs_to_update[file_type] = newest_config

        return configs_to_update, received_configs.was_successful

    def _parse_local_config(
        self, file_instance: FileInstance
    ) -> Optional[SyncVersionFacadeInterface]:
        try:
            return cast(
                SyncVersionFacadeInterface, file_instance.read_to_facade()
            )
        except RawFileError as e:
            self._report_processor.report(raw_file_error_report(e))
        except ParserErrorException as e:
            self._report_processor.report_list(
                file_instance.parser_exception_to_report_list(e)
            )
        return None

    def _parse_received_configs(
        self, file_instance: FileInstance, raw_files: Iterable[ConfigInfo]
    ) -> list[SyncVersionFacadeInterface]:
        result = []
        for file in raw_files:
            try:
                facade = cast(
                    SyncVersionFacadeInterface,
                    file_instance.raw_to_facade(
                        file.cfg_content.encode("utf-8")
                    ),
                )
                result.append(facade)
            except ParserErrorException as e:
                # Add context, so we know from which node we received invalid
                # config
                context = ReportItemContext(node=file.cfg_origin)
                report_list = [
                    replace(report, context=context)
                    for report in file_instance.parser_exception_to_report_list(
                        e, is_forced_or_warning=True
                    )
                ]
                self._report_processor.report_list(report_list)
        return result


def _file_hash(file_instance: FileInstance, facade: FacadeInterface) -> str:
    # sha1 is used to be compatible with the old ruby implementation.
    # The hash is only used to compare and sort the files, not for security
    # reasons
    return sha1(
        file_instance.facade_to_raw(facade), usedforsecurity=False
    ).hexdigest()


def _find_newest_config(
    file_instance: FileInstance, configs: Iterable[SyncVersionFacadeInterface]
) -> Optional[SyncVersionFacadeInterface]:
    if not configs:
        return None

    max_version = max(cfg.data_version for cfg in configs)
    cfg_hash = {}
    hash_count: dict[str, int] = defaultdict(int)

    for cfg in configs:
        if cfg.data_version == max_version:
            file_hash = _file_hash(file_instance, cfg)
            cfg_hash[file_hash] = cfg
            hash_count[file_hash] += 1

    # find the biggest hash among hashes with the most frequent count, so that
    # this function always returns config with the same content when receiving
    # the same amount of different configs with the same version
    most_frequent_hash_count = max(hash_count.values())
    most_frequent_hash = max(
        h for h in hash_count if hash_count[h] == most_frequent_hash_count
    )

    return cfg_hash[most_frequent_hash]



from typing import TYPE_CHECKING, Any, Mapping, Optional

from pcs.common.pacemaker.alert import CibAlertListDto
from pcs.common.types import StringIterable
from pcs.lib.cib import alert
from pcs.lib.cib.nvpair import (
    arrange_first_instance_attributes,
    arrange_first_meta_attributes,
)
from pcs.lib.cib.rule.in_effect import get_rule_evaluator
from pcs.lib.cib.tools import (
    IdProvider,
    get_alerts,
)
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError

if TYPE_CHECKING:
    from pcs.common.reports import ReportItemList


def create_alert(
    # Path is mandatory, so it should not be optional. However, the current
    # code calling this function does not prevent it being None. The idea behind
    # that was, that the validation would happen in the lib command. Since
    # then, however, the paradigma got changed as we found out that a client
    # should actually be responsible for providing all mandatory parameters.
    # The interface cannot be simply changed, as backward compatibility must be
    # maintained for lib.commands. We still want to change it, but it needs to
    # be done in the proper way.
    lib_env: LibraryEnvironment,
    alert_id: Optional[str],
    path: Optional[str],
    instance_attribute_dict: Mapping[str, str],
    meta_attribute_dict: Mapping[str, str],
    description: Optional[str] = None,
) -> None:
    """
    Create new alert.
    Raises LibraryError if path is not specified, or any other failure.

    lib_env -- LibraryEnvironment
    alert_id -- id of alert to be created, if None it will be generated
    path -- path to script for alert
    instance_attribute_dict -- dictionary of instance attributes
    meta_attribute_dict -- dictionary of meta attributes
    description -- alert description description
    """
    cib = lib_env.get_cib()
    id_provider = IdProvider(cib)

    lib_env.report_processor.report_list(
        alert.validate_create_alert(id_provider, path, alert_id)
    )
    if lib_env.report_processor.has_errors:
        raise LibraryError()

    alert_el = alert.create_alert(
        cib,
        id_provider,
        str(path),  # if path were None, validation above would raise
        alert_id,
        description,
    )
    arrange_first_instance_attributes(
        alert_el, instance_attribute_dict, id_provider
    )
    arrange_first_meta_attributes(alert_el, meta_attribute_dict, id_provider)

    lib_env.push_cib()


def update_alert(
    lib_env: LibraryEnvironment,
    alert_id: str,
    path: Optional[str],
    instance_attribute_dict: Mapping[str, str],
    meta_attribute_dict: Mapping[str, str],
    description: Optional[str] = None,
) -> None:
    """
    Update existing alert with specified id.

    lib_env -- LibraryEnvironment
    alert_id -- id of alert to be updated
    path -- new path, if None old value will stay unchanged
    instance_attribute_dict -- dictionary of instance attributes to update
    meta_attribute_dict -- dictionary of meta attributes to update
    description -- new description, if empty string, old description will be
        deleted, if None old value will stay unchanged
    """

    cib = lib_env.get_cib()
    id_provider = IdProvider(cib)
    alert_el = alert.update_alert(cib, alert_id, path, description)
    arrange_first_instance_attributes(
        alert_el, instance_attribute_dict, id_provider
    )
    arrange_first_meta_attributes(alert_el, meta_attribute_dict, id_provider)

    lib_env.push_cib()


def remove_alert(
    lib_env: LibraryEnvironment, alert_id_list: StringIterable
) -> None:
    """
    Remove alerts with specified ids.

    lib_env -- LibraryEnvironment
    alert_id_list -- list of alerts ids which should be removed
    """
    cib = lib_env.get_cib()
    report_list: ReportItemList = []
    for alert_id in alert_id_list:
        try:
            alert.remove_alert(cib, alert_id)
        except LibraryError as e:
            report_list += e.args

    if lib_env.report_processor.report_list(report_list).has_errors:
        raise LibraryError()
    lib_env.push_cib()


def add_recipient(
    # Recipient value is mandatory, so it should not be optional. However, the
    # current code calling this function does not prevent it being None. The
    # idea behind that was, that the validation would happen in the lib
    # command. Since then, however, the paradigma got changed as we found out
    # that a client should actually be responsible for providing all mandatory
    # parameters.
    # The interface cannot be simply changed, as backward compatibility must be
    # maintained for lib.commands. We still want to change it, but it needs to
    # be done in the proper way.
    lib_env: LibraryEnvironment,
    alert_id: str,
    recipient_value: Optional[str],
    instance_attribute_dict: Mapping[str, str],
    meta_attribute_dict: Mapping[str, str],
    recipient_id: Optional[str] = None,
    description: Optional[str] = None,
    allow_same_value: bool = False,
) -> None:
    """
    Add new recipient to alert witch id alert_id.

    lib_env -- LibraryEnvironment
    alert_id -- id of alert to which new recipient should be added
    recipient_value -- value of new recipient
    instance_attribute_dict -- dictionary of instance attributes to update
    meta_attribute_dict -- dictionary of meta attributes to update
    recipient_id -- id of new recipient, if None it will be generated
    description -- recipient description
    allow_same_value -- if True unique recipient value is not required
    """
    cib = lib_env.get_cib()
    id_provider = IdProvider(cib)
    alert_el = alert.find_alert(get_alerts(cib), alert_id)

    lib_env.report_processor.report_list(
        alert.validate_add_recipient(
            id_provider,
            alert_el,
            recipient_value,
            recipient_id,
            allow_same_value=allow_same_value,
        )
    )
    if lib_env.report_processor.has_errors:
        raise LibraryError()

    recipient_el = alert.add_recipient(
        id_provider,
        alert_el,
        # if recipient_value were None, validation above would raise
        str(recipient_value),
        recipient_id,
        description=description,
    )
    arrange_first_instance_attributes(
        recipient_el, instance_attribute_dict, id_provider
    )
    arrange_first_meta_attributes(
        recipient_el, meta_attribute_dict, id_provider
    )

    lib_env.push_cib()


def update_recipient(
    lib_env: LibraryEnvironment,
    recipient_id: str,
    instance_attribute_dict: Mapping[str, str],
    meta_attribute_dict: Mapping[str, str],
    recipient_value: Optional[str] = None,
    description: Optional[str] = None,
    allow_same_value: bool = False,
) -> None:
    """
    Update existing recipient.

    lib_env -- LibraryEnvironment
    recipient_id -- id of recipient to be updated
    instance_attribute_dict -- dictionary of instance attributes to update
    meta_attribute_dict -- dictionary of meta attributes to update
    recipient_value -- new recipient value, if None old value will stay
        unchanged
    description -- new description, if empty string, old description will be
        deleted, if None old value will stay unchanged
    allow_same_value -- if True unique recipient value is not required
    """
    cib = lib_env.get_cib()
    id_provider = IdProvider(cib)
    recipient_el = alert.find_recipient(get_alerts(cib), recipient_id)

    lib_env.report_processor.report_list(
        alert.validate_update_recipient(
            recipient_el,
            recipient_value,
            allow_same_value=allow_same_value,
        )
    )
    if lib_env.report_processor.has_errors:
        raise LibraryError()

    recipient = alert.update_recipient(
        recipient_el,
        recipient_value=recipient_value,
        description=description,
    )
    arrange_first_instance_attributes(
        recipient, instance_attribute_dict, id_provider
    )
    arrange_first_meta_attributes(recipient, meta_attribute_dict, id_provider)

    lib_env.push_cib()


def remove_recipient(
    lib_env: LibraryEnvironment, recipient_id_list: StringIterable
) -> None:
    """
    Remove specified recipients.

    lib_env -- LibraryEnvironment
    recipient_id_list -- list of recipients ids to be removed
    """
    cib = lib_env.get_cib()
    report_list: ReportItemList = []
    for recipient_id in recipient_id_list:
        try:
            alert.remove_recipient(cib, recipient_id)
        except LibraryError as e:
            report_list += e.args
    if lib_env.report_processor.report_list(report_list).has_errors:
        raise LibraryError()
    lib_env.push_cib()


def get_config_dto(
    lib_env: LibraryEnvironment, evaluate_expired: bool = False
) -> CibAlertListDto:
    cib = lib_env.get_cib()
    rule_in_effect_eval = get_rule_evaluator(
        cib, lib_env.cmd_runner(), lib_env.report_processor, evaluate_expired
    )
    return CibAlertListDto(
        [
            alert.alert_el_to_dto(alert_el, rule_eval=rule_in_effect_eval)
            for alert_el in alert.get_all_alert_elements(get_alerts(cib))
        ]
    )


# DEPRECATED, use get_config_dto
def get_all_alerts(lib_env: LibraryEnvironment) -> list[dict[str, Any]]:
    """
    Returns list of all alerts. See docs of pcs.lib.cib.alert.get_all_alerts for
    description of data format.

    lib_env -- LibraryEnvironment
    """
    return alert.get_all_alerts_dict(lib_env.get_cib())



from dataclasses import dataclass
from typing import (
    Optional,
    Sequence,
)

from pcs.common.const import PcmkRoleType
from pcs.common.interface.dto import DataTransferObject
from pcs.common.pacemaker.types import CibTicketLossPolicy

from .set import CibResourceSetDto


@dataclass(frozen=True)
class CibConstraintTicketAttributesDto(DataTransferObject):
    constraint_id: str
    ticket: str
    loss_policy: Optional[CibTicketLossPolicy]


@dataclass(frozen=True)
class CibConstraintTicketDto(DataTransferObject):
    resource_id: str
    role: Optional[PcmkRoleType]
    attributes: CibConstraintTicketAttributesDto


@dataclass(frozen=True)
class CibConstraintTicketSetDto(DataTransferObject):
    resource_sets: Sequence[CibResourceSetDto]
    attributes: CibConstraintTicketAttributesDto



from pcs import usage
from pcs.cli.cib.element import command as cib_element_cmd
from pcs.cli.common.routing import create_router

cib_cmd = create_router(
    {
        "element": create_router(
            {
                "description": cib_element_cmd.description,
            },
            ["cib", "element"],
        ),
        "help": lambda lib, argv, modifiers: print(usage.cib(argv)),
    },
    ["cib"],
)



from dataclasses import dataclass
from typing import (
    Optional,
    Type,
)

from pcs.common import file_type_codes as code
from pcs.common import reports
from pcs.lib.auth import config as auth_config
from pcs.lib.booth.config_facade import ConfigFacade as BoothConfigFacade
from pcs.lib.booth.config_parser import Exporter as BoothConfigExporter
from pcs.lib.booth.config_parser import Parser as BoothConfigParser
from pcs.lib.corosync.config_facade import ConfigFacade as CorosyncConfigFacade
from pcs.lib.corosync.config_parser import Exporter as CorosyncConfigExporter
from pcs.lib.corosync.config_parser import Parser as CorosyncConfigParser
from pcs.lib.dr.config.facade import Facade as DrConfigFacade
from pcs.lib.host.config.exporter import Exporter as KnownHostsExporter
from pcs.lib.host.config.facade import Facade as KnownHostsFacade
from pcs.lib.host.config.parser import Parser as KnownHostsParser
from pcs.lib.interface.config import (
    ExporterInterface,
    FacadeInterface,
    ParserErrorException,
    ParserInterface,
)
from pcs.lib.pcs_cfgsync.config.facade import Facade as CfgsyncCtlFacade
from pcs.lib.permissions import config as pcs_settings_conf

from .json import (
    JsonExporter,
    JsonParser,
)


@dataclass(frozen=True)
class FileToolbox:
    # File type code the toolbox belongs to
    file_type_code: code.FileTypeCode
    # Provides an easy access for reading and modifying data
    facade: Type[FacadeInterface]
    # Turns raw data into a structure which the facade is able to process
    parser: Type[ParserInterface]
    # Turns a structure produced by the parser and the facade to raw data
    exporter: Type[ExporterInterface]
    # Checks that the structure is valid
    validator: None  # TBI
    # Provides means for file syncing based on the file's version
    version_controller: None  # TBI


class NoopParser(ParserInterface):
    @staticmethod
    def parse(raw_file_data: bytes) -> bytes:
        return raw_file_data

    @staticmethod
    def exception_to_report_list(
        exception: ParserErrorException,
        file_type_code: code.FileTypeCode,
        file_path: Optional[str],
        force_code: Optional[reports.types.ForceCode],
        is_forced_or_warning: bool,
    ) -> reports.ReportItemList:
        del (
            exception,
            file_type_code,
            file_path,
            force_code,
            is_forced_or_warning,
        )
        return []


class NoopExporter(ExporterInterface):
    @staticmethod
    def export(config_structure: bytes) -> bytes:
        return config_structure


class NoopFacade(FacadeInterface):
    @classmethod
    def create(cls) -> "NoopFacade":
        return cls(bytes())


_toolboxes = {
    code.BOOTH_CONFIG: FileToolbox(
        file_type_code=code.BOOTH_CONFIG,
        facade=BoothConfigFacade,
        parser=BoothConfigParser,
        exporter=BoothConfigExporter,
        validator=None,  # TODO needed for files syncing
        version_controller=None,  # TODO needed for files syncing
    ),
    code.BOOTH_KEY: FileToolbox(
        file_type_code=code.BOOTH_KEY,
        facade=NoopFacade,
        parser=NoopParser,
        exporter=NoopExporter,
        validator=None,  # TODO needed for files syncing
        version_controller=None,  # TODO needed for files syncing
    ),
    code.PCS_CFGSYNC_CTL: FileToolbox(
        file_type_code=code.PCS_CFGSYNC_CTL,
        facade=CfgsyncCtlFacade,
        parser=JsonParser,
        exporter=JsonExporter,
        validator=None,
        version_controller=None,
    ),
    code.COROSYNC_CONF: FileToolbox(
        file_type_code=code.COROSYNC_CONF,
        facade=CorosyncConfigFacade,
        parser=CorosyncConfigParser,
        exporter=CorosyncConfigExporter,
        validator=None,  # TODO needed for files syncing
        version_controller=None,  # TODO needed for files syncing
    ),
    code.COROSYNC_QNETD_CA_CERT: FileToolbox(
        file_type_code=code.COROSYNC_QNETD_CA_CERT,
        facade=NoopFacade,
        parser=NoopParser,
        exporter=NoopExporter,
        validator=None,  # TODO needed for files syncing
        version_controller=None,  # TODO needed for files syncing
    ),
    code.PACEMAKER_AUTHKEY: FileToolbox(
        file_type_code=code.PACEMAKER_AUTHKEY,
        facade=NoopFacade,
        parser=NoopParser,
        exporter=NoopExporter,
        validator=None,  # TODO needed for files syncing
        version_controller=None,  # TODO needed for files syncing
    ),
    code.PCS_KNOWN_HOSTS: FileToolbox(
        file_type_code=code.PCS_KNOWN_HOSTS,
        facade=KnownHostsFacade,
        parser=KnownHostsParser,
        exporter=KnownHostsExporter,
        validator=None,  # TODO needed for files syncing
        version_controller=None,  # TODO needed for files syncing
    ),
    code.PCS_DR_CONFIG: FileToolbox(
        file_type_code=code.PCS_DR_CONFIG,
        facade=DrConfigFacade,
        parser=JsonParser,
        exporter=JsonExporter,
        validator=None,  # TODO needed for files syncing
        version_controller=None,  # TODO needed for files syncing
    ),
    code.PCS_USERS_CONF: FileToolbox(
        file_type_code=code.PCS_USERS_CONF,
        facade=auth_config.facade.Facade,
        parser=auth_config.parser.Parser,
        exporter=auth_config.exporter.Exporter,
        validator=None,  # TODO needed for files syncing
        version_controller=None,  # TODO needed for files syncing
    ),
    code.PCS_SETTINGS_CONF: FileToolbox(
        file_type_code=code.PCS_SETTINGS_CONF,
        facade=pcs_settings_conf.facade.FacadeV2,
        parser=pcs_settings_conf.parser.ParserV2,
        exporter=pcs_settings_conf.exporter.ExporterV2,
        validator=None,  # TODO needed for files syncing
        version_controller=None,  # TODO needed for files syncing
    ),
}


def for_file_type(file_type_code: code.FileTypeCode) -> FileToolbox:
    return _toolboxes[file_type_code]



from typing import (
    Optional,
    Sequence,
)

from pcs.cli.common.output import (
    INDENT_STEP,
    bool_to_cli_value,
    options_to_cmd,
    pairs_to_cmd,
)
from pcs.cli.reports.output import warn
from pcs.common.pacemaker.constraint import CibResourceSetDto
from pcs.common.str_tools import (
    format_list,
    format_optional,
    indent,
    pairs_to_text,
)
from pcs.common.types import StringSequence


def _resource_set_options_to_pairs(
    resource_set_dto: CibResourceSetDto,
) -> list[tuple[str, str]]:
    pairs = []
    if resource_set_dto.sequential is not None:
        pairs.append(
            ("sequential", bool_to_cli_value(resource_set_dto.sequential))
        )
    if resource_set_dto.require_all is not None:
        pairs.append(
            ("require-all", bool_to_cli_value(resource_set_dto.require_all))
        )
    if resource_set_dto.ordering:
        pairs.append(("ordering", resource_set_dto.ordering))
    if resource_set_dto.action:
        pairs.append(("action", resource_set_dto.action))
    if resource_set_dto.role:
        pairs.append(("role", resource_set_dto.role))
    if resource_set_dto.score:
        pairs.append(("score", resource_set_dto.score))
    if resource_set_dto.kind:
        pairs.append(("kind", resource_set_dto.kind.capitalize()))
    return pairs


def resource_set_to_text(
    resource_set_dto: CibResourceSetDto,
    with_id: bool,
) -> list[str]:
    output = [
        "Resource Set:{id}".format(
            id=f" {resource_set_dto.set_id}" if with_id else ""
        )
    ]
    set_options = [
        "Resources: {resources}".format(
            resources=format_list(resource_set_dto.resources_ids)
        )
    ] + pairs_to_text(_resource_set_options_to_pairs(resource_set_dto))
    output.extend(indent(set_options, indent_step=INDENT_STEP))
    return output


def set_constraint_to_text(
    constraint_id: str,
    constraint_attrs_lines: StringSequence,
    resource_sets: Sequence[CibResourceSetDto],
    with_id: bool,
) -> list[str]:
    header = "Set Constraint:"
    if with_id:
        header += f" {constraint_id}"
    result = [header]
    result.extend(indent(constraint_attrs_lines, indent_step=INDENT_STEP))
    for res_set_dto in resource_sets:
        result.extend(
            indent(
                resource_set_to_text(res_set_dto, with_id),
                indent_step=INDENT_STEP,
            )
        )
    return result


def resource_set_to_cmd(resource_set: CibResourceSetDto) -> Optional[list[str]]:
    filtered_pairs = []
    for pair in _resource_set_options_to_pairs(resource_set):
        # this list is based on pcs.lib.cib.constraint.resource_set._ATTRIBUTES
        if pair[0] not in ("action", "require-all", "role", "sequential"):
            warn(
                f"Option '{pair[0]}' detected in resource set "
                f"'{resource_set.set_id}' but not "
                "supported by this command."
                " Command for creating the constraint is omitted."
            )
            return None
        filtered_pairs.append(pair)

    return [
        "set {resources}{options}".format(
            resources=options_to_cmd(resource_set.resources_ids),
            options=format_optional(
                pairs_to_cmd(filtered_pairs),
                template=" {}",
            ),
        )
    ]



from unittest import TestCase, mock

from lxml import etree

from pcs.common import const, reports
from pcs.lib.cib.constraint import resource_set
from pcs.lib.cib.tools import IdProvider, Version

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_xml_equal,
)
from pcs_test.tools.custom_mock import MockLibraryReportProcessor
from pcs_test.tools.xml import etree_to_str, str_to_etree


class PrepareSetTest(TestCase):
    def setUp(self):
        self.report_processor = MockLibraryReportProcessor(debug=False)

    def test_return_corrected_resource_set(self):
        find_valid_id = mock.Mock()
        find_valid_id.side_effect = lambda id_: {"A": "AA", "B": "BB"}[id_]
        self.assertEqual(
            {"ids": ["AA", "BB"], "options": {"sequential": "true"}},
            resource_set.prepare_set(
                find_valid_id,
                {"ids": ["A", "B"], "options": {"sequential": "true"}},
                self.report_processor,
            ),
        )
        self.report_processor.assert_reports([])

    def test_refuse_invalid_attribute_name(self):
        assert_raise_library_error(
            lambda: resource_set.prepare_set(
                mock.Mock(),
                {"ids": ["A", "B"], "options": {"invalid_name": "true"}},
                self.report_processor,
            ),
        )
        self.report_processor.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["invalid_name"],
                    option_type="set",
                    allowed=["action", "require-all", "role", "sequential"],
                    allowed_patterns=[],
                ),
            ]
        )

    def test_refuse_invalid_attribute_value(self):
        assert_raise_library_error(
            lambda: resource_set.prepare_set(
                mock.Mock(),
                {"ids": ["A", "B"], "options": {"role": "invalid"}},
                self.report_processor,
            ),
        )
        self.report_processor.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="role",
                    allowed_values=const.PCMK_ROLES,
                    option_value="invalid",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ]
        )


class CreateTest(TestCase):
    def setUp(self):
        cib = str_to_etree(
            """
            <cib>
                <resources>
                    <primitive id="resource1" />
                    <primitive id="resource2" />
                </resources>
                <constraints>
                    <rsc_ticket id="my-constraint" ticket="ticket1" />
                </constraints>
            </cib>
            """
        )
        self.parent_el = cib.xpath(".//*[@id='my-constraint']")[0]
        self.id_provider = IdProvider(cib)
        self.rsc_list = ["resource1", "resource2"]

    def test_no_options(self):
        resource_set.create(
            self.parent_el,
            self.id_provider,
            Version(3, 7, 0),
            self.rsc_list,
            {},
        )
        assert_xml_equal(
            """
            <rsc_ticket id="my-constraint" ticket="ticket1">
                <resource_set id="my-constraint_set">
                    <resource_ref id="resource1"/>
                    <resource_ref id="resource2"/>
                </resource_set>
            </rsc_ticket>
            """,
            etree_to_str(self.parent_el),
        )

    def test_options(self):
        resource_set.create(
            self.parent_el,
            self.id_provider,
            Version(3, 7, 0),
            self.rsc_list,
            {
                "id": "my-set",
                "role": const.PCMK_ROLE_PROMOTED,
                "option": "value",
                "empty": "",
            },
        )
        assert_xml_equal(
            f"""
            <rsc_ticket id="my-constraint" ticket="ticket1">
                <resource_set id="my-set" option="value"
                    role="{const.PCMK_ROLE_PROMOTED}"
                >
                    <resource_ref id="resource1"/>
                    <resource_ref id="resource2"/>
                </resource_set>
            </rsc_ticket>
            """,
            etree_to_str(self.parent_el),
        )

    def test_legacy_role(self):
        resource_set.create(
            self.parent_el,
            self.id_provider,
            Version(3, 6, 0),
            self.rsc_list,
            {"role": const.PCMK_ROLE_PROMOTED},
        )
        assert_xml_equal(
            f"""
            <rsc_ticket id="my-constraint" ticket="ticket1">
                <resource_set id="my-constraint_set"
                    role="{const.PCMK_ROLE_PROMOTED_LEGACY}"
                >
                    <resource_ref id="resource1"/>
                    <resource_ref id="resource2"/>
                </resource_set>
            </rsc_ticket>
            """,
            etree_to_str(self.parent_el),
        )


class CreateOldTest(TestCase):
    def test_resource_set_to_parent(self):
        # pylint: disable=no-self-use
        constraint_element = etree.Element("constraint")
        resource_set.create_old(
            constraint_element,
            {"ids": ["A", "B"], "options": {"sequential": "true"}},
        )
        assert_xml_equal(
            etree.tostring(constraint_element).decode(),
            """
            <constraint>
              <resource_set id="constraint_set_set" sequential="true">
                <resource_ref id="A"></resource_ref>
                <resource_ref id="B"></resource_ref>
              </resource_set>
            </constraint>
        """,
        )


class GetResourceIdListTest(TestCase):
    def test_returns_id_list_from_element(self):
        element = etree.Element("resource_set")
        for _id in ("A", "B"):
            etree.SubElement(element, "resource_ref").attrib["id"] = _id

        self.assertEqual(
            ["A", "B"], resource_set.get_resource_id_set_list(element)
        )



from unittest import TestCase

from pcs.cli.stonith.levels import output
from pcs.common.pacemaker.fencing_topology import (
    CibFencingLevelAttributeDto,
    CibFencingLevelNodeDto,
    CibFencingLevelRegexDto,
    CibFencingTopologyDto,
)

FIXTURE_TARGET_NODE_DTO_LIST = [
    CibFencingLevelNodeDto("fl1", "node1", 1, ["d1"]),
    CibFencingLevelNodeDto("fl2", "node1", 2, ["d2"]),
    CibFencingLevelNodeDto("fl3", "node2", 3, ["d3", "d4"]),
]
FIXTURE_REGEX_DTO_LIST = [
    CibFencingLevelRegexDto("fl1", "node.*", 1, ["d1"]),
    CibFencingLevelRegexDto("fl2", "node.*", 2, ["d2"]),
    CibFencingLevelRegexDto("fl3", ".*node", 3, ["d3", "d4"]),
]
FIXTURE_ATTRIBUTE_DTO_LIST = [
    CibFencingLevelAttributeDto("fl1", "A", "B", 1, ["d1"]),
    CibFencingLevelAttributeDto("fl2", "A", "B", 2, ["d2"]),
    CibFencingLevelAttributeDto("fl3", "A", "X", 2, ["d2"]),
    CibFencingLevelAttributeDto("fl4", "B", "B", 3, ["d3", "d4"]),
]


class LevelsToOutputMixin:
    def test_empty(self):
        command_output = self._call_command(CibFencingTopologyDto([], [], []))
        self.assertEqual(command_output, [])

    def test_nodes(self):
        command_output = self._call_command(
            CibFencingTopologyDto(FIXTURE_TARGET_NODE_DTO_LIST, [], [])
        )
        self.assertEqual(command_output, self.NODE_OUTPUT)

    def test_regex(self):
        command_output = self._call_command(
            CibFencingTopologyDto([], FIXTURE_REGEX_DTO_LIST, [])
        )
        self.assertEqual(command_output, self.REGEX_OUTPUT)

    def test_attributes(self):
        command_output = self._call_command(
            CibFencingTopologyDto([], [], FIXTURE_ATTRIBUTE_DTO_LIST)
        )
        self.assertEqual(command_output, self.ATTRIBUTE_OUTPUT)

    def test_combination(self):
        command_output = self._call_command(
            CibFencingTopologyDto(
                FIXTURE_TARGET_NODE_DTO_LIST,
                FIXTURE_REGEX_DTO_LIST,
                FIXTURE_ATTRIBUTE_DTO_LIST,
            )
        )
        self.assertEqual(
            command_output,
            self.NODE_OUTPUT + self.REGEX_OUTPUT + self.ATTRIBUTE_OUTPUT,
        )


class LevelsToText(TestCase, LevelsToOutputMixin):
    NODE_OUTPUT = [
        "Target (node): node1",
        "  Level 1: d1",
        "  Level 2: d2",
        "Target (node): node2",
        "  Level 3: d3 d4",
    ]

    REGEX_OUTPUT = [
        "Target (regexp): .*node",
        "  Level 3: d3 d4",
        "Target (regexp): node.*",
        "  Level 1: d1",
        "  Level 2: d2",
    ]

    ATTRIBUTE_OUTPUT = [
        "Target (attribute): A=B",
        "  Level 1: d1",
        "  Level 2: d2",
        "Target (attribute): B=B",
        "  Level 3: d3 d4",
        "Target (attribute): A=X",
        "  Level 2: d2",
    ]

    def _call_command(self, dto):
        # pylint: disable=no-self-use
        return output.stonith_level_config_to_text(dto)


class LevelsToCmd(TestCase, LevelsToOutputMixin):
    NODE_OUTPUT = [
        "pcs stonith level add --force -- 1 node1 d1 id=fl1",
        "pcs stonith level add --force -- 2 node1 d2 id=fl2",
        "pcs stonith level add --force -- 3 node2 d3 d4 id=fl3",
    ]

    REGEX_OUTPUT = [
        "pcs stonith level add --force -- 1 regexp%node.* d1 id=fl1",
        "pcs stonith level add --force -- 2 regexp%node.* d2 id=fl2",
        "pcs stonith level add --force -- 3 regexp%.*node d3 d4 id=fl3",
    ]

    ATTRIBUTE_OUTPUT = [
        "pcs stonith level add --force -- 1 attrib%A=B d1 id=fl1",
        "pcs stonith level add --force -- 2 attrib%A=B d2 id=fl2",
        "pcs stonith level add --force -- 2 attrib%A=X d2 id=fl3",
        "pcs stonith level add --force -- 3 attrib%B=B d3 d4 id=fl4",
    ]

    def _call_command(self, dto):
        # pylint: disable=no-self-use
        return output.stonith_level_config_to_cmd(dto)



from unittest import TestCase

from pcs.common import reports
from pcs.lib.cib.resource import meta as lib_meta
from pcs.lib.resource_agent import ResourceAgentParameter
from pcs.lib.resource_agent import const as ra_const

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal
from pcs_test.tools.metadata_dto import (
    FIXTURE_KNOWN_META_NAMES_PRIMITIVE_META,
    get_fixture_meta_attributes_dto,
)

FIXTURE_METADATA_PARAMETERS = [
    ResourceAgentParameter(
        name=parameter_dto.name,
        shortdesc=parameter_dto.shortdesc,
        longdesc=parameter_dto.longdesc,
        type=type,
        default=parameter_dto.default,
        enum_values=parameter_dto.enum_values,
        required=parameter_dto.required,
        advanced=parameter_dto.advanced,
        deprecated=parameter_dto.deprecated,
        deprecated_by=parameter_dto.deprecated_by,
        deprecated_desc=parameter_dto.deprecated_desc,
        unique_group=parameter_dto.unique_group,
        reloadable=parameter_dto.reloadable,
    )
    for parameter_dto in get_fixture_meta_attributes_dto().parameters
]


class TestValidateMetaAttributes(TestCase):
    meta_types = [ra_const.STONITH_META, ra_const.PRIMITIVE_META]

    def test_empty(self):
        assert_report_item_list_equal(
            lib_meta.validate_meta_attributes(
                self.meta_types, FIXTURE_METADATA_PARAMETERS, {}
            ),
            [],
        )

    def test_known_meta(self):
        meta_attrs = {
            "priority": "INFINITY",
            "critical": "true",
            "target-role": "Stopped",
            "failure-timeout": "10s",
            "remote-node": "some-remote-node",
            "remote-port": "1234",
            "remote-connect-timeout": "60s",
        }
        assert_report_item_list_equal(
            lib_meta.validate_meta_attributes(
                self.meta_types, FIXTURE_METADATA_PARAMETERS, meta_attrs
            ),
            [],
        )

    def test_unknown_meta(self):
        meta_attrs = {
            "target_role": "Stopped",
            "non-existent-name": "10s",
            "remote-pot": "1234",
        }
        unknown_meta = sorted(meta_attrs.keys())
        assert_report_item_list_equal(
            lib_meta.validate_meta_attributes(
                self.meta_types, FIXTURE_METADATA_PARAMETERS, meta_attrs
            ),
            [
                fixture.warn(
                    reports.codes.META_ATTRS_UNKNOWN_TO_PCMK,
                    unknown_meta=unknown_meta,
                    known_meta=FIXTURE_KNOWN_META_NAMES_PRIMITIVE_META,
                    meta_types=sorted(self.meta_types),
                )
            ],
        )

    def test_mixed_known_unknown_meta(self):
        meta_attrs = {
            "target-role": "Stopped",
            "non-existent-name": "10s",
            "remote-pot": "",
        }
        unknown_meta = ["non-existent-name"]
        assert_report_item_list_equal(
            lib_meta.validate_meta_attributes(
                self.meta_types, FIXTURE_METADATA_PARAMETERS, meta_attrs
            ),
            [
                fixture.warn(
                    reports.codes.META_ATTRS_UNKNOWN_TO_PCMK,
                    unknown_meta=unknown_meta,
                    known_meta=FIXTURE_KNOWN_META_NAMES_PRIMITIVE_META,
                    meta_types=sorted(self.meta_types),
                )
            ],
        )

    def test_meta_with_empty_values(self):
        meta_attrs = {
            "target-role": "",
            "non-existent-name": "",
            "remote-pot": "",
        }
        assert_report_item_list_equal(
            lib_meta.validate_meta_attributes(
                self.meta_types, FIXTURE_METADATA_PARAMETERS, meta_attrs
            ),
            [],
        )



try:
    from pcs.daemon.app import webui
except ImportError:
    # You need to skip tests in tests that uses AppTest in the case webui
    # is not there.
    webui = None

from pcs_test.tier0.daemon.app import fixtures_app


class AppTest(fixtures_app.AppTest):
    def setUp(self):
        self.session_storage = webui.session.Storage(lifetime_seconds=10)
        super().setUp()

    def fetch(self, path, raise_error=False, **kwargs):
        if "sid" in kwargs:
            if "headers" not in kwargs:
                kwargs["headers"] = {}
            kwargs["headers"]["Cookie"] = (
                f"{webui.auth.PCSD_SESSION}={kwargs['sid']}"
            )
            del kwargs["sid"]

        return super().fetch(path, raise_error=raise_error, **kwargs)

    def create_login_session(self):
        return self.session_storage.login(fixtures_app.USER)



import dataclasses
from typing import (
    Any,
    Optional,
    cast,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    Argv,
    InputModifiers,
)
from pcs.cli.constraint import command
from pcs.cli.constraint.output import print_config
from pcs.cli.constraint_ticket import parse_args
from pcs.cli.reports.output import error
from pcs.cli.reports.preprocessor import (
    get_duplicate_constraint_exists_preprocessor,
)
from pcs.common import reports
from pcs.common.pacemaker.constraint import CibConstraintsDto
from pcs.common.reports.messages import InvalidOptions


def create_with_set(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    create ticket constraint with resource set
    object lib exposes library
    list argv see usage for "constraint colocation set"
    dict like object modifiers can contain
        "force" allows resource in clone/master and constraint duplicity

    Options:
      * --force - allow resource inside clone (or master), allow duplicate
        element
      * -f - CIB file
    """
    modifiers.ensure_only_supported("--force", "-f")
    lib.env.report_processor.set_report_item_preprocessor(
        get_duplicate_constraint_exists_preprocessor(lib)
    )
    command.create_with_set(
        lib.constraint_ticket.create_with_set,
        argv,
        modifiers,
    )


def add(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    create ticket constraint
    object lib exposes library
    list argv see usage for "constraint colocation add"
    dict like object modifiers can contain
        "force" allows resource in clone/master and constraint duplicity

    Options:
      * --force - allow resource inside clone (or master), allow duplicate
        element
      * -f - CIB file
    """

    generic_preprocessor = get_duplicate_constraint_exists_preprocessor(lib)

    def _rsc_role_preprocessor(
        report_item: reports.ReportItem,
    ) -> Optional[reports.ReportItem]:
        if isinstance(report_item.message, InvalidOptions):
            new_message = dataclasses.replace(
                report_item.message,
                allowed=sorted(set(report_item.message.allowed) - {"rsc-role"}),
            )
            return dataclasses.replace(report_item, message=new_message)
        return report_item

    def _report_item_preprocessor(
        report_item: reports.ReportItem,
    ) -> Optional[reports.ReportItem]:
        report_item_2 = generic_preprocessor(report_item)
        if not report_item_2:
            return None
        return _rsc_role_preprocessor(report_item_2)

    modifiers.ensure_only_supported("--force", "-f")
    ticket, resource_id, resource_role, options = parse_args.parse_add(argv)
    if "rsc-role" in options:
        raise CmdLineInputError(
            "Resource role must not be specified among options"
            + ", specify it before resource id"
        )

    if resource_role:
        options["rsc-role"] = resource_role

    lib.env.report_processor.set_report_item_preprocessor(
        _report_item_preprocessor
    )

    lib.constraint_ticket.create(
        ticket,
        resource_id,
        options,
        resource_in_clone_alowed=modifiers.get("--force"),
        duplication_alowed=modifiers.get("--force"),
    )


def remove(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    if len(argv) != 2:
        raise CmdLineInputError()
    ticket, resource_id = argv
    if not lib.constraint_ticket.remove(ticket, resource_id):
        raise error("no matching ticket constraint found")


def config_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    modifiers.ensure_only_supported("-f", "--output-format", "--full")
    if argv:
        raise CmdLineInputError()

    constraints_dto = cast(
        CibConstraintsDto,
        lib.constraint.get_config(evaluate_rules=True),
    )

    print_config(
        CibConstraintsDto(
            ticket=constraints_dto.ticket,
            ticket_set=constraints_dto.ticket_set,
        ),
        modifiers,
    )



from unittest import TestCase, mock

from pcs.cli.cib.element import command
from pcs.cli.common.errors import CmdLineInputError

from pcs_test.tools.misc import dict_to_modifiers


class Description(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cib"])
        self.lib.cib = mock.Mock(
            spec_set=["element_description_get", "element_description_set"]
        )

    def _call_cmd(self, argv=None):
        command.description(self.lib, argv or [], dict_to_modifiers({}))

    def test_success_set(self):
        self._call_cmd(["a", "description"])
        self.lib.cib.element_description_set.assert_called_once_with(
            "a", "description"
        )
        self.lib.cib.element_description_get.assert_not_called()

    @mock.patch("pcs.cli.cib.element.command.print")
    def test_success_get(self, mock_print):
        self.lib.cib.element_description_get.return_value = "AAA"
        self._call_cmd(["a"])
        self.lib.cib.element_description_set.assert_not_called()
        self.lib.cib.element_description_get.assert_called_once_with("a")
        mock_print.assert_called_once_with("AAA")

    @mock.patch("pcs.cli.cib.element.command.print")
    def test_success_get_no_description(self, mock_print):
        self.lib.cib.element_description_get.return_value = ""
        self._call_cmd(["a"])
        self.lib.cib.element_description_set.assert_not_called()
        self.lib.cib.element_description_get.assert_called_once_with("a")
        mock_print.assert_not_called()

    def test_bad_number_of_args(self):
        bad_args = [
            [],
            ["a", "b", "c"],
        ]
        for argv in bad_args:
            with self.subTest(argv=argv):
                with self.assertRaises(CmdLineInputError) as cm:
                    self._call_cmd(argv)
                self.assertIsNone(cm.exception.message)
                self.lib.cib.element_description_set.assert_not_called()
                self.lib.cib.element_description_get.assert_not_called()



import logging
from unittest import TestCase, mock

import pcs.lib.commands.alert as cmd_alert
from pcs.common.pacemaker.alert import (
    CibAlertDto,
    CibAlertListDto,
    CibAlertRecipientDto,
    CibAlertSelectDto,
)
from pcs.common.pacemaker.nvset import CibNvpairDto, CibNvsetDto
from pcs.common.pacemaker.rule import CibRuleExpressionDto
from pcs.common.reports import ReportItemSeverity as Severities
from pcs.common.reports import codes as report_codes
from pcs.common.types import CibRuleExpressionType, CibRuleInEffectStatus
from pcs.lib.cib.rule.in_effect import RuleInEffectEval
from pcs.lib.env import LibraryEnvironment

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.custom_mock import (
    MockLibraryReportProcessor,
    RuleInEffectEvalMock,
)


class CreateAlertTest(TestCase):
    fixture_final_alerts = """
        <alerts>
            <alert id="my-alert" path="/my/path" description="my description">
                <meta_attributes id="my-alert-meta_attributes">
                    <nvpair
                        id="my-alert-meta_attributes-meta1"
                        name="meta1"
                        value="val1"
                    />
                </meta_attributes>
                <instance_attributes id="my-alert-instance_attributes">
                    <nvpair
                        id="my-alert-instance_attributes-another"
                        name="another"
                        value="val"
                    />
                    <nvpair
                        id="my-alert-instance_attributes-instance"
                        name="instance"
                        value="value"
                    />
                </instance_attributes>
            </alert>
        </alerts>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_validation(self):
        self.config.runner.cib.load()
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.create_alert(
                self.env_assist.get_env(), "1alert", None, None, None
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                    option_names=["path"],
                    option_type=None,
                ),
                fixture.error(
                    report_codes.INVALID_ID_BAD_CHAR,
                    id="1alert",
                    id_description="alert-id",
                    invalid_character="1",
                    is_first_char=True,
                ),
            ],
        )

    def test_create_no_upgrade(self):
        self.config.runner.cib.load()
        self.config.env.push_cib(optional_in_conf=self.fixture_final_alerts)
        cmd_alert.create_alert(
            self.env_assist.get_env(),
            "my-alert",
            "/my/path",
            {"instance": "value", "another": "val"},
            {"meta1": "val1"},
            "my description",
        )


class UpdateAlertTest(TestCase):
    fixture_initial_alerts = """
        <alerts>
            <alert id="my-alert" path="/my/path" description="my description">
                <instance_attributes id="my-alert-instance_attributes">
                    <nvpair
                        id="my-alert-instance_attributes-instance"
                        name="instance"
                        value="value"
                    />
                    <nvpair
                        id="my-alert-instance_attributes-another"
                        name="another"
                        value="val"
                    />
                </instance_attributes>
                <meta_attributes id="my-alert-meta_attributes">
                    <nvpair
                        id="my-alert-meta_attributes-meta1"
                        name="meta1"
                        value="val1"
                    />
                </meta_attributes>
            </alert>
        </alerts>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_update_all(self):
        fixture_final_alerts = """
        <alerts>
            <alert id="my-alert" path="/another/one">
                <instance_attributes id="my-alert-instance_attributes">
                    <nvpair
                        id="my-alert-instance_attributes-another"
                        name="another"
                        value="val"
                    />
                    <nvpair
                        id="my-alert-instance_attributes-my-attr"
                        name="my-attr"
                        value="its_val"
                    />
                </instance_attributes>
                <meta_attributes id="my-alert-meta_attributes">
                    <nvpair
                        id="my-alert-meta_attributes-meta1"
                        name="meta1"
                        value="val2"
                    />
                </meta_attributes>
            </alert>
        </alerts>
        """
        (
            self.config.runner.cib.load(
                optional_in_conf=self.fixture_initial_alerts
            ).env.push_cib(
                replace={"./configuration/alerts": fixture_final_alerts}
            )
        )
        cmd_alert.update_alert(
            self.env_assist.get_env(),
            "my-alert",
            "/another/one",
            {"instance": "", "my-attr": "its_val"},
            {"meta1": "val2"},
            "",
        )

    def test_update_instance_attribute(self):
        (
            self.config.runner.cib.load(
                optional_in_conf=self.fixture_initial_alerts
            ).env.push_cib(
                replace={
                    './configuration/alerts/alert[@id="my-alert"]/'
                    'instance_attributes/nvpair[@name="instance"]': """
                        <nvpair
                            id="my-alert-instance_attributes-instance"
                            name="instance"
                            value="new_val"
                        />
                    """
                }
            )
        )
        cmd_alert.update_alert(
            self.env_assist.get_env(),
            "my-alert",
            None,
            {"instance": "new_val"},
            {},
            None,
        )

    def test_alert_doesnt_exist(self):
        (
            self.config.runner.cib.load(
                optional_in_conf="""
                    <alerts>
                        <alert id="alert" path="path"/>
                    </alerts>
                """
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.update_alert(
                self.env_assist.get_env(), "unknown", "test", {}, {}, None
            ),
            [
                (
                    Severities.ERROR,
                    report_codes.ID_NOT_FOUND,
                    {
                        "context_type": "alerts",
                        "context_id": "",
                        "id": "unknown",
                        "expected_types": ["alert"],
                    },
                    None,
                ),
            ],
            expected_in_processor=False,
        )


class RemoveAlertTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.cib.load(
            optional_in_conf="""
                <alerts>
                    <alert id="alert1" path="path"/>
                    <alert id="alert2" path="/path"/>
                    <alert id="alert3" path="/path"/>
                    <alert id="alert4" path="/path"/>
                </alerts>
            """
        )

    def test_one_alert(self):
        self.config.env.push_cib(
            remove="./configuration/alerts/alert[@id='alert2']"
        )
        cmd_alert.remove_alert(self.env_assist.get_env(), ["alert2"])

    def test_multiple_alerts(self):
        self.config.env.push_cib(
            remove=[
                "./configuration/alerts/alert[@id='alert1']",
                "./configuration/alerts/alert[@id='alert3']",
                "./configuration/alerts/alert[@id='alert4']",
            ]
        )
        cmd_alert.remove_alert(
            self.env_assist.get_env(), ["alert1", "alert3", "alert4"]
        )

    def test_no_alert(self):
        self.config.env.push_cib()
        cmd_alert.remove_alert(self.env_assist.get_env(), [])

    def test_alerts_dont_exist(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.remove_alert(
                self.env_assist.get_env(),
                ["unknown1", "alert1", "unknown2", "alert2"],
            )
        )
        self.env_assist.assert_reports(
            [
                (
                    Severities.ERROR,
                    report_codes.ID_NOT_FOUND,
                    {
                        "context_type": "alerts",
                        "context_id": "",
                        "id": "unknown1",
                        "expected_types": ["alert"],
                    },
                    None,
                ),
                (
                    Severities.ERROR,
                    report_codes.ID_NOT_FOUND,
                    {
                        "context_type": "alerts",
                        "context_id": "",
                        "id": "unknown2",
                        "expected_types": ["alert"],
                    },
                    None,
                ),
            ]
        )


class AddRecipientTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.cib.load(
            optional_in_conf="""
                <alerts>
                    <alert id="alert" path="path">
                        <recipient id="alert-recipient" value="value1"/>
                    </alert>
                </alerts>
            """
        )

    def test_alert_doesnt_exist(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.add_recipient(
                self.env_assist.get_env(), "wrong-id", "value1", {}, {}
            ),
            [
                fixture.error(
                    report_codes.ID_NOT_FOUND,
                    context_type="alerts",
                    context_id="",
                    id="wrong-id",
                    expected_types=["alert"],
                )
            ],
            expected_in_processor=False,
        )

    def test_value_not_defined(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.add_recipient(
                self.env_assist.get_env(), "alert", "", {}, {}
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                    option_names=["value"],
                    option_type=None,
                )
            ],
        )

    def test_recipient_already_exists(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.add_recipient(
                self.env_assist.get_env(),
                "alert",
                "value1",
                {},
                {},
                recipient_id="alert-recipient",
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.ID_ALREADY_EXISTS,
                    id="alert-recipient",
                )
            ],
        )

    def test_duplicity_of_value_forced(self):
        self.config.env.push_cib(
            replace={
                './/alert[@id="alert"]': """
                    <alert id="alert" path="path">
                        <recipient id="alert-recipient" value="value1"/>
                        <recipient id="alert-recipient-1" value="value1"/>
                    </alert>
                """
            }
        )
        cmd_alert.add_recipient(
            self.env_assist.get_env(),
            "alert",
            "value1",
            {},
            {},
            allow_same_value=True,
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.CIB_ALERT_RECIPIENT_ALREADY_EXISTS,
                    alert="alert",
                    recipient="value1",
                )
            ]
        )

    def test_without_id(self):
        self.config.env.push_cib(
            replace={
                './/alert[@id="alert"]': """
                <alert id="alert" path="path">
                    <recipient id="alert-recipient" value="value1"/>
                    <recipient id="alert-recipient-1" value="value">
                        <meta_attributes
                            id="alert-recipient-1-meta_attributes"
                        >
                            <nvpair
                                id="alert-recipient-1-meta_attributes-attr1"
                                name="attr1"
                                value="val1"
                            />
                            <nvpair
                                id="alert-recipient-1-meta_attributes-attr2"
                                name="attr2"
                                value="val2"
                            />
                        </meta_attributes>
                        <instance_attributes
                            id="alert-recipient-1-instance_attributes"
                        >
                            <nvpair
                                id="alert-recipient-1-instance_attributes-attr1"
                                name="attr1"
                                value="val1"
                            />
                        </instance_attributes>
                    </recipient>
                </alert>
                """
            }
        )
        cmd_alert.add_recipient(
            self.env_assist.get_env(),
            "alert",
            "value",
            {"attr1": "val1"},
            {"attr2": "val2", "attr1": "val1"},
        )

    def test_with_id(self):
        self.config.env.push_cib(
            replace={
                './/alert[@id="alert"]': """
                <alert id="alert" path="path">
                    <recipient id="alert-recipient" value="value1"/>
                    <recipient id="my-recipient" value="value">
                        <meta_attributes
                            id="my-recipient-meta_attributes"
                        >
                            <nvpair
                                id="my-recipient-meta_attributes-attr1"
                                name="attr1"
                                value="val1"
                            />
                            <nvpair
                                id="my-recipient-meta_attributes-attr2"
                                name="attr2"
                                value="val2"
                            />
                        </meta_attributes>
                        <instance_attributes
                            id="my-recipient-instance_attributes"
                        >
                            <nvpair
                                id="my-recipient-instance_attributes-attr1"
                                name="attr1"
                                value="val1"
                            />
                        </instance_attributes>
                    </recipient>
                </alert>
                """
            }
        )
        cmd_alert.add_recipient(
            self.env_assist.get_env(),
            "alert",
            "value",
            {"attr1": "val1"},
            {"attr2": "val2", "attr1": "val1"},
            recipient_id="my-recipient",
        )


class UpdateRecipientTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.cib.load(
            optional_in_conf="""
                <alerts>
                    <alert id="alert" path="path">
                        <recipient id="alert-recipient" value="value1"/>
                        <recipient id="alert-recipient-1" value="value"
                            description="d"
                        >
                            <meta_attributes
                                id="alert-recipient-1-meta_attributes"
                            >
                                <nvpair
                                    id="alert-recipient-1-meta_attributes-attr1"
                                    name="attr1"
                                    value="val1"
                                />
                                <nvpair
                                    id="alert-recipient-1-meta_attributes-attr2"
                                    name="attr2"
                                    value="val2"
                                />
                            </meta_attributes>
                            <instance_attributes
                                id="alert-recipient-1-instance_attributes"
                            >
                                <nvpair
                                    id="alert-recipient-1-instance_attributes-attr1"
                                    name="attr1"
                                    value="val1"
                                />
                            </instance_attributes>
                        </recipient>
                    </alert>
                </alerts>
            """
        )

    def test_empty_value(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.update_recipient(
                self.env_assist.get_env(),
                "alert-recipient-1",
                {},
                {},
                recipient_value="",
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.CIB_ALERT_RECIPIENT_VALUE_INVALID,
                    recipient="",
                )
            ],
        )

    def test_recipient_not_found(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.update_recipient(
                self.env_assist.get_env(), "recipient", {}, {}
            ),
            [
                fixture.error(
                    report_codes.ID_NOT_FOUND,
                    id="recipient",
                    expected_types=["recipient"],
                    context_id="",
                    context_type="alerts",
                )
            ],
            expected_in_processor=False,
        )

    def test_update_duplicity_allowed(self):
        self.config.env.push_cib(
            replace={
                './/alert[@id="alert"]': """
                <alert id="alert" path="path">
                    <recipient id="alert-recipient" value="value1"/>
                    <recipient id="alert-recipient-1" value="value1"
                        description="d"
                    >
                        <meta_attributes
                            id="alert-recipient-1-meta_attributes"
                        >
                            <nvpair
                                id="alert-recipient-1-meta_attributes-attr1"
                                name="attr1"
                                value="val1"
                            />
                            <nvpair
                                id="alert-recipient-1-meta_attributes-attr2"
                                name="attr2"
                                value="val2"
                            />
                        </meta_attributes>
                        <instance_attributes
                            id="alert-recipient-1-instance_attributes"
                        >
                            <nvpair
                                id="alert-recipient-1-instance_attributes-attr1"
                                name="attr1"
                                value="val1"
                            />
                        </instance_attributes>
                    </recipient>
                </alert>
                """,
            }
        )
        cmd_alert.update_recipient(
            self.env_assist.get_env(),
            "alert-recipient-1",
            {},
            {},
            recipient_value="value1",
            description=None,
            allow_same_value=True,
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.CIB_ALERT_RECIPIENT_ALREADY_EXISTS,
                    alert="alert",
                    recipient="value1",
                )
            ]
        )

    def test_update_all(self):
        self.config.env.push_cib(
            replace={
                './/alert[@id="alert"]': """
                <alert id="alert" path="path">
                    <recipient id="alert-recipient" value="value1"/>
                    <recipient
                        id="alert-recipient-1"
                        value="new_val"
                        description="desc"
                    >
                        <meta_attributes
                            id="alert-recipient-1-meta_attributes"
                        >
                            <nvpair
                                id="alert-recipient-1-meta_attributes-attr2"
                                name="attr2"
                                value="val2"
                            />
                            <nvpair
                                id="alert-recipient-1-meta_attributes-attr3"
                                name="attr3"
                                value="new_val"
                            />
                        </meta_attributes>
                        <instance_attributes
                            id="alert-recipient-1-instance_attributes"
                        >
                            <nvpair
                                id="alert-recipient-1-instance_attributes-attr1"
                                name="attr1"
                                value="value"
                            />
                        </instance_attributes>
                    </recipient>
                </alert>
                """,
            }
        )
        cmd_alert.update_recipient(
            self.env_assist.get_env(),
            "alert-recipient-1",
            {"attr1": "value"},
            {"attr1": "", "attr3": "new_val"},
            recipient_value="new_val",
            description="desc",
        )


class RemoveRecipientTest(TestCase):
    fixture_initial_alerts = """
        <alerts>
            <alert id="alert" path="path">
                <recipient id="alert-recip1" value="value1"/>
                <recipient id="alert-recip2" value="value2"/>
            </alert>
            <alert id="alert2" path="path">
                <recipient id="alert2-recip3" value="value3"/>
                <recipient id="alert2-recip4" value="value4"/>
            </alert>
        </alerts>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.cib.load(
            optional_in_conf=self.fixture_initial_alerts
        )

    def test_recipient_not_found(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_alert.remove_recipient(
                self.env_assist.get_env(),
                ["recipient", "alert-recip1", "alert2-recip1"],
            )
        )
        self.env_assist.assert_reports(
            [
                (
                    Severities.ERROR,
                    report_codes.ID_NOT_FOUND,
                    {
                        "id": "recipient",
                        "expected_types": ["recipient"],
                        "context_type": "alerts",
                        "context_id": "",
                    },
                    None,
                ),
                (
                    Severities.ERROR,
                    report_codes.ID_NOT_FOUND,
                    {
                        "id": "alert2-recip1",
                        "expected_types": ["recipient"],
                        "context_type": "alerts",
                        "context_id": "",
                    },
                    None,
                ),
            ]
        )

    def test_one_recipient(self):
        self.config.env.push_cib(
            remove="./configuration/alerts/alert/recipient[@id='alert-recip1']"
        )
        cmd_alert.remove_recipient(self.env_assist.get_env(), ["alert-recip1"])

    def test_multiple_recipients(self):
        self.config.env.push_cib(
            remove=[
                "./configuration/alerts/alert/recipient[@id='alert-recip1']",
                "./configuration/alerts/alert/recipient[@id='alert-recip2']",
                "./configuration/alerts/alert/recipient[@id='alert2-recip4']",
            ]
        )
        cmd_alert.remove_recipient(
            self.env_assist.get_env(),
            ["alert-recip1", "alert-recip2", "alert2-recip4"],
        )

    def test_no_recipient(self):
        self.config.env.push_cib()
        cmd_alert.remove_recipient(self.env_assist.get_env(), [])


class GetConfigDto(TestCase):
    fixture_alerts = """
        <alerts>
            <alert id="alert-all" path="/path/all" description="all options">
                <recipient id="alert-all-recipient" value="value-all"
                    description="all options recipient"
                >
                    <instance_attributes>
                        <nvpair id="alert-all-recipient-ia"
                            name="all-iar1-name" value="all-iar1-value"
                        />
                        <rule id="alert-all-recipient-ia-rule">
                            <date_expression
                                id="alert-all-recipient-ia-rule-de"
                                operation="gt" end="2000-01-01"
                            />
                        </rule>
                    </instance_attributes>
                    <meta_attributes>
                        <nvpair id="alert-all-recipient-ma"
                            name="all-mar1-name" value="all-mar1-value"
                        />
                        <rule id="alert-all-recipient-ma-rule" boolean-op="and">
                            <date_expression id="alert-all-recipient-ma-rule-de"
                            operation="lt" end="2000-01-01" />
                        </rule>
                    </meta_attributes>
                </recipient>
                <instance_attributes>
                    <nvpair id="alert-all-ia"
                        name="all-iaa1-name" value="all-iaa1-value"
                    />
                    <rule id="alert-all-ia-rule">
                        <date_expression id="alert-all-ia-rule-de"
                            operation="gt" end="2000-01-01"
                        />
                    </rule>
                </instance_attributes>
                <meta_attributes>
                    <nvpair id="alert-all-ma"
                        name="all-maa1-name" value="all-maa1-value"
                    />
                    <rule id="alert-all-ma-rule" boolean-op="and">
                        <date_expression id="alert-all-ma-rule-de"
                        operation="lt" end="2000-01-01" />
                    </rule>
                </meta_attributes>
                <select>
                    <select_nodes />
                    <select_attributes />
                </select>
            </alert>
        </alerts>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def get_alerts_dto(self, rule_eval: RuleInEffectEval) -> CibAlertListDto:
        return CibAlertListDto(
            [
                CibAlertDto(
                    id="alert-all",
                    path="/path/all",
                    description="all options",
                    recipients=[
                        CibAlertRecipientDto(
                            id="alert-all-recipient",
                            value="value-all",
                            description="all options recipient",
                            meta_attributes=[
                                CibNvsetDto(
                                    id="",
                                    options={},
                                    rule=CibRuleExpressionDto(
                                        id="alert-all-recipient-ma-rule",
                                        type=CibRuleExpressionType.RULE,
                                        in_effect=rule_eval.get_rule_status(
                                            "alert-all-recipient-ma-rule"
                                        ),
                                        options={"boolean-op": "and"},
                                        date_spec=None,
                                        duration=None,
                                        expressions=[
                                            CibRuleExpressionDto(
                                                id="alert-all-recipient-ma-rule-de",
                                                type=CibRuleExpressionType.DATE_EXPRESSION,
                                                in_effect=CibRuleInEffectStatus.UNKNOWN,
                                                options={
                                                    "operation": "lt",
                                                    "end": "2000-01-01",
                                                },
                                                date_spec=None,
                                                duration=None,
                                                expressions=[],
                                                as_string="date lt 2000-01-01",
                                            )
                                        ],
                                        as_string="date lt 2000-01-01",
                                    ),
                                    nvpairs=[
                                        CibNvpairDto(
                                            id="alert-all-recipient-ma",
                                            name="all-mar1-name",
                                            value="all-mar1-value",
                                        )
                                    ],
                                )
                            ],
                            instance_attributes=[
                                CibNvsetDto(
                                    id="",
                                    options={},
                                    rule=CibRuleExpressionDto(
                                        id="alert-all-recipient-ia-rule",
                                        type=CibRuleExpressionType.RULE,
                                        in_effect=rule_eval.get_rule_status(
                                            "alert-all-recipient-ia-rule"
                                        ),
                                        options={},
                                        date_spec=None,
                                        duration=None,
                                        expressions=[
                                            CibRuleExpressionDto(
                                                id="alert-all-recipient-ia-rule-de",
                                                type=CibRuleExpressionType.DATE_EXPRESSION,
                                                in_effect=CibRuleInEffectStatus.UNKNOWN,
                                                options={
                                                    "operation": "gt",
                                                    "end": "2000-01-01",
                                                },
                                                date_spec=None,
                                                duration=None,
                                                expressions=[],
                                                as_string="date gt 2000-01-01",
                                            )
                                        ],
                                        as_string="date gt 2000-01-01",
                                    ),
                                    nvpairs=[
                                        CibNvpairDto(
                                            id="alert-all-recipient-ia",
                                            name="all-iar1-name",
                                            value="all-iar1-value",
                                        )
                                    ],
                                )
                            ],
                        )
                    ],
                    select=CibAlertSelectDto(
                        nodes=True,
                        fencing=False,
                        resources=False,
                        attributes=True,
                        attributes_select=[],
                    ),
                    meta_attributes=[
                        CibNvsetDto(
                            id="",
                            options={},
                            rule=CibRuleExpressionDto(
                                id="alert-all-ma-rule",
                                type=CibRuleExpressionType.RULE,
                                in_effect=rule_eval.get_rule_status(
                                    "alert-all-ma-rule"
                                ),
                                options={"boolean-op": "and"},
                                date_spec=None,
                                duration=None,
                                expressions=[
                                    CibRuleExpressionDto(
                                        id="alert-all-ma-rule-de",
                                        type=CibRuleExpressionType.DATE_EXPRESSION,
                                        in_effect=CibRuleInEffectStatus.UNKNOWN,
                                        options={
                                            "operation": "lt",
                                            "end": "2000-01-01",
                                        },
                                        date_spec=None,
                                        duration=None,
                                        expressions=[],
                                        as_string="date lt 2000-01-01",
                                    )
                                ],
                                as_string="date lt 2000-01-01",
                            ),
                            nvpairs=[
                                CibNvpairDto(
                                    id="alert-all-ma",
                                    name="all-maa1-name",
                                    value="all-maa1-value",
                                )
                            ],
                        )
                    ],
                    instance_attributes=[
                        CibNvsetDto(
                            id="",
                            options={},
                            rule=CibRuleExpressionDto(
                                id="alert-all-ia-rule",
                                type=CibRuleExpressionType.RULE,
                                in_effect=rule_eval.get_rule_status(
                                    "alert-all-ia-rule"
                                ),
                                options={},
                                date_spec=None,
                                duration=None,
                                expressions=[
                                    CibRuleExpressionDto(
                                        id="alert-all-ia-rule-de",
                                        type=CibRuleExpressionType.DATE_EXPRESSION,
                                        in_effect=CibRuleInEffectStatus.UNKNOWN,
                                        options={
                                            "operation": "gt",
                                            "end": "2000-01-01",
                                        },
                                        date_spec=None,
                                        duration=None,
                                        expressions=[],
                                        as_string="date gt 2000-01-01",
                                    )
                                ],
                                as_string="date gt 2000-01-01",
                            ),
                            nvpairs=[
                                CibNvpairDto(
                                    id="alert-all-ia",
                                    name="all-iaa1-name",
                                    value="all-iaa1-value",
                                )
                            ],
                        )
                    ],
                ),
            ]
        )

    def test_success_no_alerts(self):
        self.config.runner.cib.load()
        self.assertEqual(
            cmd_alert.get_config_dto(self.env_assist.get_env()),
            CibAlertListDto([]),
        )

    @mock.patch("pcs.lib.commands.alert.get_rule_evaluator")
    def test_success(self, mock_get_rule_evaluator):
        self.config.runner.cib.load(optional_in_conf=self.fixture_alerts)
        rule_evaluator = RuleInEffectEvalMock(
            {
                "alert-all-recipient-ia-rule": CibRuleInEffectStatus.IN_EFFECT,
                "alert-all-recipient-ma-rule": CibRuleInEffectStatus.EXPIRED,
                "alert-all-ma-rule": CibRuleInEffectStatus.EXPIRED,
                "alert-all-ia-rule": CibRuleInEffectStatus.IN_EFFECT,
            }
        )
        mock_get_rule_evaluator.return_value = rule_evaluator
        self.assertEqual(
            cmd_alert.get_config_dto(
                self.env_assist.get_env(), evaluate_expired=True
            ),
            self.get_alerts_dto(rule_evaluator),
        )
        mock_get_rule_evaluator.assert_called_once()

    @mock.patch("pcs.lib.cib.rule.in_effect.has_rule_in_effect_status_tool")
    def test_success_no_rule_evaluation(self, mock_has_rule_tool):
        mock_has_rule_tool.side_effect = AssertionError(
            "has_rule_in_effect_status_tool should not be called"
        )
        self.config.runner.cib.load(optional_in_conf=self.fixture_alerts)
        self.assertEqual(
            cmd_alert.get_config_dto(
                self.env_assist.get_env(), evaluate_expired=False
            ),
            self.get_alerts_dto(RuleInEffectEvalMock({})),
        )


@mock.patch.object(LibraryEnvironment, "get_cib")
@mock.patch("pcs.lib.cib.alert.get_all_alerts_dict")
class GetAllAlertsTest(TestCase):
    def setUp(self):
        self.mock_log = mock.MagicMock(spec_set=logging.Logger)
        self.mock_rep = MockLibraryReportProcessor()
        self.mock_env = LibraryEnvironment(self.mock_log, self.mock_rep)

    def test_success(self, mock_alerts, mock_get_cib):
        mock_alerts.return_value = [{"id": "alert"}]
        mock_get_cib.return_value = "<cib/>"
        self.assertEqual(
            [{"id": "alert"}], cmd_alert.get_all_alerts(self.mock_env)
        )
        self.assertEqual(1, mock_alerts.call_count)



from unittest import (
    TestCase,
    mock,
)

from lxml import etree

from pcs import settings
from pcs.lib.pacemaker import api_result

from pcs_test.tools import fixture_crm_mon
from pcs_test.tools.assertions import assert_xml_equal
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.xml import etree_to_str


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_rng/api/api-result.rng")
)
class GetApiResultDom(TestCase):
    def test_valid_xml(self):
        # pylint: disable=no-self-use
        xml = """
            <pacemaker-result api-version="2.3" request="command">
                <status code="0" message="OK"/>
            </pacemaker-result>
        """
        result_el = api_result.get_api_result_dom(xml)
        assert_xml_equal(xml, etree_to_str(result_el))

    def test_syntax_error_xml(self):
        xml = "<syntax_error>"
        self.assertRaises(
            etree.XMLSyntaxError, lambda: api_result.get_api_result_dom(xml)
        )

    def test_invalid_xml(self):
        xml = "<pacemaker-result/>"
        self.assertRaises(
            etree.DocumentInvalid, lambda: api_result.get_api_result_dom(xml)
        )


class GetStatusFromApiResult(TestCase):
    # pylint: disable=protected-access
    def test_errors(self):
        self.assertEqual(
            api_result.get_status_from_api_result(
                etree.fromstring(
                    fixture_crm_mon.error_xml(
                        123, "short message", ["error1", "error2"]
                    )
                )
            ),
            api_result.Status(123, "short message", ["error1", "error2"]),
        )

    def test_no_errors(self):
        self.assertEqual(
            api_result.get_status_from_api_result(
                etree.fromstring(
                    fixture_crm_mon.error_xml(123, "short message")
                )
            ),
            api_result.Status(123, "short message", []),
        )



from unittest import (
    TestCase,
    mock,
)

import pcs.lib.commands.acl as cmd_acl
from pcs.common import reports
from pcs.lib.env import LibraryEnvironment

from pcs_test.tools import fixture
from pcs_test.tools.assertions import ExtendedAssertionsMixin
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.custom_mock import MockLibraryReportProcessor


class AclCommandsTest(TestCase, ExtendedAssertionsMixin):
    def setUp(self):
        self.mock_rep = MockLibraryReportProcessor()
        self.mock_env = mock.MagicMock(spec_set=LibraryEnvironment)
        self.mock_env.report_processor = self.mock_rep
        self.cib = "cib"
        self.mock_env.get_cib.return_value = self.cib

    def assert_get_cib_called(self):
        self.mock_env.get_cib.assert_called_once_with()

    def assert_same_cib_pushed(self):
        self.mock_env.push_cib.assert_called_once_with()

    def assert_cib_not_pushed(self):
        self.assertEqual(0, self.mock_env.push_cib.call_count)


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
class CibAclSection(TestCase):
    def test_push_cib_on_success(self):
        # pylint: disable=no-self-use
        env = mock.MagicMock()
        env.get_cib = mock.Mock(return_value="cib")
        with cmd_acl.cib_acl_section(env):
            pass
        env.get_cib.assert_called_once_with()
        env.push_cib.assert_called_once_with()

    def test_does_not_push_cib_on_exception(self):
        env = mock.MagicMock()

        def run():
            with cmd_acl.cib_acl_section(env):
                raise AssertionError()

        self.assertRaises(AssertionError, run)
        env.get_cib.assert_called_once_with()
        env.push_cib.assert_not_called()


class CreateRoleTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.runner.cib.load(
            resources="""
            <resources>
                <primitive id="R1" class="ocf" provider="pacemaker" type="Dummy"/>
                <primitive id="R2" class="ocf" provider="pacemaker" type="Dummy"/>
            </resources>
            """
        )

    def test_success_just_role(self):
        acl_xml = """
            <acls>
                <acl_role description="desc" id="role_id" />
            </acls>
        """
        self.config.env.push_cib(optional_in_conf=acl_xml)

        cmd_acl.create_role(self.env_assist.get_env(), "role_id", [], "desc")

    def test_success_with_permissions(self):
        acl_xml = """
            <acls>
                <acl_role id="role_id">
                    <acl_permission id="role_id-write" kind="write" reference="R1" />
                    <acl_permission id="role_id-read" kind="read" xpath="./" />
                </acl_role>
            </acls>
        """
        self.config.env.push_cib(optional_in_conf=acl_xml)

        cmd_acl.create_role(
            self.env_assist.get_env(),
            "role_id",
            [("write", "id", "R1"), ("read", "xpath", "./")],
            "",
        )

    def test_id_already_used(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_acl.create_role(self.env_assist.get_env(), "R1", [], "")
        )
        self.env_assist.assert_reports(
            [
                fixture.error(reports.codes.ID_ALREADY_EXISTS, id="R1"),
            ]
        )

    def test_validation(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_acl.create_role(
                self.env_assist.get_env(),
                "#role_id",
                [
                    ("change", "id", "R1"),
                    ("read", "path", "./"),
                    ("deny", "id", "R3"),
                ],
                "",
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_ID_BAD_CHAR,
                    id="#role_id",
                    id_description="ACL role",
                    invalid_character="#",
                    is_first_char=True,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="permission",
                    option_value="change",
                    allowed_values=["read", "write", "deny"],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="scope type",
                    option_value="path",
                    allowed_values=["xpath", "id"],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id="R3",
                    expected_types=["id"],
                    context_type="",
                    context_id="",
                ),
            ]
        )


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.remove_role")
class RemoveRoleTest(AclCommandsTest):
    def test_success(self, mock_remove):
        cmd_acl.remove_role(self.mock_env, "role_id", False)
        self.assert_get_cib_called()
        mock_remove.assert_called_once_with(self.cib, "role_id", False)
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.find_target_or_group")
@mock.patch("pcs.lib.cib.acl.assign_role")
class AssignRoleNotSpecific(AclCommandsTest, ExtendedAssertionsMixin):
    def test_success(self, mock_assign, find_target_or_group):
        find_target_or_group.return_value = "target_el"
        cmd_acl.assign_role_not_specific(self.mock_env, "role_id", "target_id")
        self.assert_get_cib_called()
        find_target_or_group.assert_called_once_with(self.cib, "target_id")
        mock_assign.assert_called_once_with(self.cib, "role_id", "target_el")
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.find_target")
@mock.patch("pcs.lib.cib.acl.assign_role")
class AssignRoleToTargetTest(AclCommandsTest):
    def test_success(self, mock_assign, find_target):
        find_target.return_value = "target_el"
        cmd_acl.assign_role_to_target(self.mock_env, "role_id", "target_id")
        self.assert_get_cib_called()
        mock_assign.assert_called_once_with(self.cib, "role_id", "target_el")
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.find_group")
@mock.patch("pcs.lib.cib.acl.assign_role")
class AssignRoleToGroupTest(AclCommandsTest):
    def test_success(self, mock_assign, find_group):
        find_group.return_value = "group_el"
        cmd_acl.assign_role_to_group(self.mock_env, "role_id", "group_id")
        self.assert_get_cib_called()
        mock_assign.assert_called_once_with(self.cib, "role_id", "group_el")
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.unassign_role")
@mock.patch("pcs.lib.cib.acl.find_target_or_group")
class UnassignRoleNotSpecificTest(AclCommandsTest):
    def test_success(self, find_target_or_group, mock_unassign):
        find_target_or_group.return_value = "target_el"
        cmd_acl.unassign_role_not_specific(
            self.mock_env, "role_id", "target_id", False
        )
        self.assert_get_cib_called()
        find_target_or_group.assert_called_once_with(self.cib, "target_id")
        mock_unassign.assert_called_once_with("target_el", "role_id", False)
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.unassign_role")
@mock.patch("pcs.lib.cib.acl.find_target")
class UnassignRoleFromTargetTest(AclCommandsTest):
    def test_success(self, find_target, mock_unassign):
        find_target.return_value = "el"
        cmd_acl.unassign_role_from_target(
            self.mock_env, "role_id", "el_id", False
        )
        self.assert_get_cib_called()
        find_target.assert_called_once_with(self.cib, "el_id")
        mock_unassign.assert_called_once_with("el", "role_id", False)
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.unassign_role")
@mock.patch("pcs.lib.cib.acl.find_group")
class UnassignRoleFromGroupTest(AclCommandsTest):
    def test_success(self, find_group, mock_unassign):
        find_group.return_value = "el"
        cmd_acl.unassign_role_from_group(
            self.mock_env, "role_id", "el_id", False
        )
        self.assert_get_cib_called()
        find_group.assert_called_once_with(self.cib, "el_id")
        mock_unassign.assert_called_once_with("el", "role_id", False)
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.create_target")
@mock.patch("pcs.lib.cib.acl.assign_all_roles")
class CreateTargetTest(AclCommandsTest):
    def test_success(self, mock_assign, mock_create):
        mock_create.return_value = "el"
        cmd_acl.create_target(
            self.mock_env, "el_id", ["role1", "role2", "role3"]
        )
        self.assert_get_cib_called()
        mock_create.assert_called_once_with(self.cib, "el_id")
        mock_assign(self.cib, "el", ["role1", "role2", "role3"])
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.create_group")
@mock.patch("pcs.lib.cib.acl.assign_all_roles")
class CreateGroupTest(AclCommandsTest):
    def test_success(self, mock_assign, mock_create):
        mock_create.return_value = "el"
        cmd_acl.create_group(
            self.mock_env, "el_id", ["role1", "role2", "role3"]
        )
        self.assert_get_cib_called()
        mock_create.assert_called_once_with(self.cib, "el_id")
        mock_assign(self.cib, "el", ["role1", "role2", "role3"])
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.remove_target")
class RemoveTargetTest(AclCommandsTest):
    def test_success(self, mock_remove):
        cmd_acl.remove_target(self.mock_env, "el_id")
        self.assert_get_cib_called()
        mock_remove.assert_called_once_with(self.cib, "el_id")
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.remove_group")
class RemoveGroupTest(AclCommandsTest):
    def test_success(self, mock_remove):
        cmd_acl.remove_group(self.mock_env, "el_id")
        self.assert_get_cib_called()
        mock_remove.assert_called_once_with(self.cib, "el_id")
        self.assert_same_cib_pushed()


class AddPermissionTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.runner.cib.load(
            resources="""
            <resources>
                <primitive id="R1" class="ocf" provider="pacemaker" type="Dummy"/>
                <primitive id="R2" class="ocf" provider="pacemaker" type="Dummy"/>
            </resources>
            """,
            optional_in_conf=self.fixture_acls_xml(),
        )

    @staticmethod
    def fixture_acls_xml(roles="", permissions=""):
        return f"""
            <acls>
                <acl_role id="role1">
                    <acl_permission id="role1-write" kind="write" reference="R1" />
                    <acl_permission id="role1-read" kind="read" xpath="/" />
                    {permissions}
                </acl_role>
                {roles}
            </acls>
            """

    def test_existing_role(self):
        self.config.env.push_cib(
            optional_in_conf=self.fixture_acls_xml(
                permissions="""
                  <acl_permission id="role1-deny" kind="deny" reference="R2" />
                  <acl_permission id="role1-write-1" kind="write" xpath="/" />
                """
            )
        )
        cmd_acl.add_permission(
            self.env_assist.get_env(),
            "role1",
            [("deny", "id", "R2"), ("write", "xpath", "/")],
        )

    def test_nonexisting_role(self):
        self.config.env.push_cib(
            optional_in_conf=self.fixture_acls_xml(
                roles="""
                <acl_role id="role2">
                  <acl_permission id="role2-deny" kind="deny" reference="R2" />
                  <acl_permission id="role2-write" kind="write" xpath="/" />
                </acl_role>
                """
            )
        )
        cmd_acl.add_permission(
            self.env_assist.get_env(),
            "role2",
            [("deny", "id", "R2"), ("write", "xpath", "/")],
        )

    def test_not_role(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_acl.add_permission(self.env_assist.get_env(), "R1", []),
            [
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="R1",
                    expected_types=["acl_role"],
                    current_type="primitive",
                ),
            ],
            expected_in_processor=False,
        )

    def test_validation(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_acl.add_permission(
                self.env_assist.get_env(),
                "#role_id",
                [
                    ("change", "id", "R1"),
                    ("read", "path", "./"),
                    ("deny", "id", "R3"),
                ],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_ID_BAD_CHAR,
                    id="#role_id",
                    id_description="ACL role",
                    invalid_character="#",
                    is_first_char=True,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="permission",
                    option_value="change",
                    allowed_values=["read", "write", "deny"],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="scope type",
                    option_value="path",
                    allowed_values=["xpath", "id"],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id="R3",
                    expected_types=["id"],
                    context_type="",
                    context_id="",
                ),
            ]
        )


@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
@mock.patch("pcs.lib.cib.acl.remove_permission")
class RemovePermission(AclCommandsTest):
    def test_success(self, mock_remove):
        cmd_acl.remove_permission(self.mock_env, "id")
        self.assert_get_cib_called()
        mock_remove.assert_called_once_with(self.cib, "id")
        self.assert_same_cib_pushed()


@mock.patch("pcs.lib.cib.acl.get_target_list")
@mock.patch("pcs.lib.cib.acl.get_group_list")
@mock.patch("pcs.lib.cib.acl.get_role_list")
@mock.patch("pcs.lib.commands.acl.get_acls", mock.Mock(side_effect=lambda x: x))
class GetConfigTest(AclCommandsTest):
    def test_success(self, mock_role, mock_group, mock_target):
        mock_role.return_value = "role"
        mock_group.return_value = "group"
        mock_target.return_value = "target"
        self.assertEqual(
            {
                "target_list": "target",
                "group_list": "group",
                "role_list": "role",
            },
            cmd_acl.get_config(self.mock_env),
        )



from typing import Any

from pcs import (
    stonith,
    utils,
)
from pcs.cli.cluster_property.output import PropertyConfigurationFacade
from pcs.cli.common import parse_args
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    ArgsByKeywords,
    Argv,
    InputModifiers,
    KeyValueParser,
)
from pcs.cli.common.tools import print_to_stderr
from pcs.cli.reports import process_library_reports
from pcs.common.str_tools import (
    format_list,
    indent,
)
from pcs.lib.node import get_existing_nodes_names
from pcs.lib.pacemaker.values import is_false


def quorum_config_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --corosync_conf - mocked corosync configuration file
    """
    modifiers.ensure_only_supported("--corosync_conf")
    if argv:
        raise CmdLineInputError()
    lines = quorum_config_to_str(lib.quorum.get_config())
    if lines:
        print("\n".join(lines))


def quorum_config_to_str(config):
    """
    Commandline options: no options
    """
    lines = []

    if "options" in config and config["options"]:
        lines.append("Options:")
        lines.extend(
            indent(
                [
                    f"{name}: {value}"
                    for name, value in sorted(config["options"].items())
                ]
            )
        )

    if "device" in config and config["device"]:
        lines.append("Device:")
        lines.extend(
            indent(
                [
                    f"{name}: {value}"
                    for name, value in sorted(
                        config["device"].get("generic_options", {}).items()
                    )
                ]
            )
        )

        model_settings = [
            "Model: {m}".format(m=config["device"].get("model", ""))
        ]
        model_settings.extend(
            indent(
                [
                    f"{name}: {value}"
                    for name, value in sorted(
                        config["device"].get("model_options", {}).items()
                    )
                ]
            )
        )
        lines.extend(indent(model_settings))

        heuristics_options = config["device"].get("heuristics_options", {})
        if heuristics_options:
            heuristics_settings = ["Heuristics:"]
            heuristics_settings.extend(
                indent(
                    [
                        f"{name}: {value}"
                        for name, value in sorted(heuristics_options.items())
                    ]
                )
            )
            lines.extend(indent(heuristics_settings))

    return lines


def quorum_expected_votes_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if len(argv) != 1:
        raise CmdLineInputError()
    lib.quorum.set_expected_votes_live(argv[0])


def quorum_status_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if argv:
        raise CmdLineInputError()
    print(lib.quorum.status_text())


def quorum_update_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --skip-offline - skip offline nodes
      * --force - force changes
      * --corosync_conf - mocked corosync configuration file
      * --request-timeout - HTTP timeout, has effect only if --corosync_conf is
        not specified
    """
    modifiers.ensure_only_supported(
        "--skip-offline", "--force", "--corosync_conf", "--request-timeout"
    )
    options = KeyValueParser(argv).get_unique()
    if not options:
        raise CmdLineInputError()

    lib.quorum.set_options(
        options,
        skip_offline_nodes=modifiers.get("--skip-offline"),
        force=modifiers.get("--force"),
    )


def _parse_quorum_device_groups(arg_list: Argv) -> ArgsByKeywords:
    keyword_list = ["model", "heuristics"]
    groups = parse_args.group_by_keywords(
        arg_list, set(keyword_list), implicit_first_keyword="generic"
    )
    groups.ensure_unique_keywords()
    for keyword in keyword_list:
        if not groups.has_keyword(keyword):
            continue
        if not groups.get_args_flat(keyword):
            raise CmdLineInputError(f"No {keyword} options specified")
    return groups


def quorum_device_add_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --force - allow unknown model or options
      * --skip-offline - skip offline nodes
      * --request-timeout - HTTP timeout, has effect only if --corosync_conf is
        not specified
      * --corosync_conf - mocked corosync configuration file
    """
    modifiers.ensure_only_supported(
        "--force", "--skip-offline", "--request-timeout", "--corosync_conf"
    )
    groups = _parse_quorum_device_groups(argv)
    model_and_model_options = groups.get_args_flat("model")
    # we expect "model" keyword once, followed by the actual model value
    if not model_and_model_options or "=" in model_and_model_options[0]:
        raise CmdLineInputError()

    generic_options = KeyValueParser(
        groups.get_args_flat("generic")
    ).get_unique()
    model = model_and_model_options[0]
    model_options = KeyValueParser(model_and_model_options[1:]).get_unique()
    heuristics_options = KeyValueParser(
        groups.get_args_flat("heuristics")
    ).get_unique()

    if "model" in generic_options:
        raise CmdLineInputError("Model cannot be specified in generic options")

    lib.quorum.add_device(
        model,
        model_options,
        generic_options,
        heuristics_options,
        force_model=modifiers.get("--force"),
        force_options=modifiers.get("--force"),
        skip_offline_nodes=modifiers.get("--skip-offline"),
    )


def quorum_device_remove_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --skip-offline - skip offline nodes
      * --corosync_conf - mocked corosync configuration file
      * --request-timeout - HTTP timeout, has effect only if --corosync_conf is
        not specified
    """
    modifiers.ensure_only_supported(
        "--skip-offline", "--request-timeout", "--corosync_conf"
    )
    if argv:
        raise CmdLineInputError()

    lib.quorum.remove_device(skip_offline_nodes=modifiers.get("--skip-offline"))


def quorum_device_status_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --full - more detailed output
    """
    modifiers.ensure_only_supported("--full")
    if argv:
        raise CmdLineInputError()
    print(lib.quorum.status_device_text(modifiers.get("--full")))


def quorum_device_update_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --force - allow unknown options
      * --skip-offline - skip offline nodes
      * --corosync_conf - mocked corosync configuration file
      * --request-timeout - HTTP timeout, has effect only if --corosync_conf is
        not specified
    """
    modifiers.ensure_only_supported(
        "--force", "--skip-offline", "--request-timeout", "--corosync_conf"
    )
    groups = _parse_quorum_device_groups(argv)
    if groups.is_empty():
        raise CmdLineInputError()
    generic_options = KeyValueParser(
        groups.get_args_flat("generic")
    ).get_unique()
    model_options = KeyValueParser(groups.get_args_flat("model")).get_unique()
    heuristics_options = KeyValueParser(
        groups.get_args_flat("heuristics")
    ).get_unique()

    if "model" in generic_options:
        raise CmdLineInputError("Model cannot be specified in generic options")

    lib.quorum.update_device(
        model_options,
        generic_options,
        heuristics_options,
        force_options=modifiers.get("--force"),
        skip_offline_nodes=modifiers.get("--skip-offline"),
    )


def quorum_device_heuristics_remove_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --skip-offline - skip offline nodes
      * --corosync_conf - mocked corosync configuration file
      * --request-timeout - HTTP timeout, has effect only if --corosync_conf is
        not specified
    """
    modifiers.ensure_only_supported(
        "--skip-offline", "--corosync_conf", "--request-timeout"
    )
    if argv:
        raise CmdLineInputError()
    lib.quorum.remove_device_heuristics(
        skip_offline_nodes=modifiers.get("--skip-offline"),
    )


# TODO switch to new architecture, move to lib
def quorum_unblock_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --force - required for unblocking the quorum - DEPRECATED
      * --yes - required for unblocking the quorum
    """
    modifiers.ensure_only_supported("--force", "--yes")
    if argv:
        raise CmdLineInputError()

    output, retval = utils.run(
        ["corosync-cmapctl", "-g", "runtime.votequorum.wait_for_all_status"]
    )
    if (retval == 1 and "Error CS_ERR_NOT_EXIST" in output) or (
        retval == 0 and output.rsplit("=", maxsplit=1)[-1].strip() != "1"
    ):
        utils.err("cluster is not waiting for nodes to establish quorum")
    if retval != 0:
        utils.err("unable to check quorum status")

    all_nodes, report_list = get_existing_nodes_names(
        utils.get_corosync_conf_facade()
    )
    if report_list:
        process_library_reports(report_list)

    unjoined_nodes = list(set(all_nodes) - set(utils.getCorosyncActiveNodes()))
    if not unjoined_nodes:
        utils.err("no unjoined nodes found")
    if not utils.get_continue_confirmation(
        f"If node(s) {format_list(unjoined_nodes)} are not powered off or they "
        "do have access to shared resources, data corruption and/or cluster "
        "failure may occur",
        bool(modifiers.get("--yes")),
        bool(modifiers.get("--force")),
    ):
        return
    for node in unjoined_nodes:
        # pass --yes so no warning will be displayed
        stonith.stonith_confirm(
            lib, [node], parse_args.InputModifiers({"--yes": ""})
        )

    output, retval = utils.run(
        ["corosync-cmapctl", "-s", "quorum.cancel_wait_for_all", "u8", "1"]
    )
    if retval != 0:
        utils.err("unable to cancel waiting for nodes")
    print_to_stderr("Quorum unblocked")

    properties_facade = PropertyConfigurationFacade.from_properties_config(
        lib.cluster_property.get_properties(),
    )
    startup_fencing = str(
        properties_facade.get_property_value("startup-fencing", "")
    )
    lib.cluster_property.set_properties(
        {
            "startup-fencing": (
                "false" if not is_false(startup_fencing) else "true"
            )
        }
    )
    lib.cluster_property.set_properties({"startup-fencing": startup_fencing})
    print_to_stderr("Waiting for nodes canceled")


def check_local_qnetd_certs_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    modifiers.ensure_only_supported()
    if not argv or len(argv) != 2 or not argv[0] or not argv[1]:
        raise CmdLineInputError(
            "Expected arguments: <qnetd_host> <cluster_name>"
        )
    qnetd_host = argv[0]
    cluster_name = argv[1]
    result = lib.quorum.device_net_certificate_check_local(
        qnetd_host, cluster_name
    )
    if result:
        print("certificate present")
    else:
        print("certificate missing")


def setup_local_qnetd_certs_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    modifiers.ensure_only_supported()
    if not argv or len(argv) != 2 or not argv[0] or not argv[1]:
        raise CmdLineInputError(
            "Expected arguments: <qnetd_host> <cluster_name>"
        )
    qnetd_host = argv[0]
    cluster_name = argv[1]
    lib.quorum.device_net_certificate_setup_local(qnetd_host, cluster_name)



from unittest import TestCase, mock

from pcs import settings
from pcs.common import reports
from pcs.lib.commands import cluster_property

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import get_test_resource as rc

from .crm_attribute_mixins import (
    CrmAttributeLoadMetadataMixin,
    CrmAttributeMetadataErrorMixin,
)
from .fixtures import fixture_crm_config_properties

ALLOWED_PROPERTIES = [
    "batch-limit",
    "cluster-delay",
    "cluster-ipc-limit",
    "cluster-recheck-interval",
    "concurrent-fencing",
    "dc-deadtime",
    "election-timeout",
    "enable-acl",
    "enable-startup-probes",
    "fence-reaction",
    "fencing-enabled",
    "join-finalization-timeout",
    "join-integration-timeout",
    "load-threshold",
    "maintenance-mode",
    "migration-limit",
    "no-quorum-policy",
    "node-action-limit",
    "node-health-base",
    "node-health-green",
    "node-health-red",
    "node-health-strategy",
    "node-health-yellow",
    "node-pending-timeout",
    "pe-error-series-max",
    "pe-input-series-max",
    "pe-warn-series-max",
    "placement-strategy",
    "priority-fencing-delay",
    "remove-after-stop",
    "shutdown-escalation",
    "shutdown-lock",
    "shutdown-lock-limit",
    "start-failure-is-fatal",
    "startup-fencing",
    "stonith-action",
    "stonith-enabled",
    "stonith-max-attempts",
    "stonith-timeout",
    "stonith-watchdog-timeout",
    "stop-all-resources",
    "stop-orphan-actions",
    "stop-orphan-resources",
    "symmetric-cluster",
    "transition-delay",
]


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_rng/api/api-result.rng"),
)
class TestPropertySetCrmAttribute(
    CrmAttributeLoadMetadataMixin,
    CrmAttributeMetadataErrorMixin,
    TestCase,
):
    _load_cib_when_metadata_error = True

    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def command(self, prop_dict, force_codes=None):
        cluster_property.set_properties(
            self.env_assist.get_env(),
            prop_dict,
            [] if force_codes is None else force_codes,
        )

    def metadata_error_command(self):
        return self.command({}, [])

    def test_no_properties_specified(self):
        self.config.runner.cib.load()
        self.load_fake_agent_metadata()
        self.env_assist.assert_raise_library_error(lambda: self.command({}))
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_ITEMS_NOT_SPECIFIED,
                    force_code=reports.codes.FORCE,
                    container_type="property_set",
                    item_type="property",
                    container_id="cib-bootstrap-options",
                )
            ]
        )

    def test_no_properties_specified_forced(self):
        self.config.runner.cib.load()
        self.load_fake_agent_metadata()
        self.config.env.push_cib(
            crm_config=fixture_crm_config_properties(
                [("cib-bootstrap-options", {})]
            )
        )

        self.command({}, [reports.codes.FORCE])
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.ADD_REMOVE_ITEMS_NOT_SPECIFIED,
                    container_type="property_set",
                    item_type="property",
                    container_id="cib-bootstrap-options",
                )
            ]
        )

    def test_no_sets_and_cib_bootstrap_options_id_taken(self):
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive class="ocf" id="cib-bootstrap-options"
                        provider="pacemaker" type="Dummy"/>
                </resources>
            """
        )
        self.env_assist.assert_raise_library_error(
            lambda: self.command({"no-quorum-policy": "freeze"}),
            reports=[
                fixture.error(
                    reports.codes.CANNOT_CREATE_DEFAULT_CLUSTER_PROPERTY_SET,
                    nvset_id="cib-bootstrap-options",
                )
            ],
            expected_in_processor=False,
        )
        self.env_assist.assert_reports([])

    def test_create_cib_bootstrap_options(self):
        self.config.runner.cib.load()
        self.load_fake_agent_metadata()
        self.config.env.push_cib(
            crm_config=fixture_crm_config_properties(
                [("cib-bootstrap-options", {"maintenance-mode": "false"})]
            )
        )

        cluster_property.set_properties(
            self.env_assist.get_env(),
            {"maintenance-mode": "false"},
            [],
        )
        self.env_assist.assert_reports([])

    def _readonly_properties(self, force):
        self.config.runner.cib.load(
            crm_config=fixture_crm_config_properties(
                [
                    (
                        "cib-bootstrap-options",
                        {
                            "cluster-infrastructure": "corosync",
                            "cluster-name": "ClusterName",
                            "dc-version": "2.1.4-5.el9-dc6eb4362e",
                        },
                    )
                ]
            )
        )
        self.load_fake_agent_metadata()
        self.env_assist.assert_raise_library_error(
            lambda: cluster_property.set_properties(
                self.env_assist.get_env(),
                {
                    "cluster-name": "HACluster",
                    "dc-version": "",
                    "have-watchdog": "yes",
                },
                [reports.codes.FORCE] if force else [],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=[
                        "cluster-name",
                        "dc-version",
                        "have-watchdog",
                    ],
                    allowed=ALLOWED_PROPERTIES,
                    option_type="cluster property",
                    allowed_patterns=[],
                )
            ]
        )

    def test_readonly_properties(self):
        self._readonly_properties(False)

    def test_readonly_properties_forced(self):
        self._readonly_properties(True)

    def test_success(self):
        orig_properties = {
            "batch-limit": "10",
            "enable-acl": "true",
            "stonith-enabled": "true",
        }
        new_properties = {
            "enable-acl": "false",
            "stonith-enabled": "true",
            "maintenance-mode": "false",
        }

        self.config.runner.cib.load(
            crm_config=fixture_crm_config_properties(
                [
                    ("first-set", orig_properties),
                    ("second-set", orig_properties),
                ]
            )
        )
        self.load_fake_agent_metadata()
        self.config.env.push_cib(
            crm_config=fixture_crm_config_properties(
                [
                    ("first-set", new_properties),
                    ("second-set", orig_properties),
                ]
            )
        )

        cluster_property.set_properties(
            self.env_assist.get_env(),
            {
                "batch-limit": "",
                "enable-acl": "false",
                "maintenance-mode": "false",
            },
            [],
        )
        self.env_assist.assert_reports([])

    def test_validator_errors(self):
        self.config.runner.cib.load(
            crm_config=fixture_crm_config_properties(
                [
                    ("cib-bootstrap-options", {}),
                ]
            )
        )
        self.load_fake_agent_metadata()

        self.env_assist.assert_raise_library_error(
            lambda: cluster_property.set_properties(
                self.env_assist.get_env(),
                {
                    "unknown": "property",
                    "enable-acl": "Falsch",
                    "batch-limit": "",
                    "non-existing": "",
                },
                [],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    force_code=reports.codes.FORCE,
                    option_names=["unknown"],
                    allowed=ALLOWED_PROPERTIES,
                    option_type="cluster property",
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    force_code=reports.codes.FORCE,
                    option_name="enable-acl",
                    option_value="Falsch",
                    allowed_values=(
                        "a pacemaker boolean value: '0', '1', 'false', 'n', "
                        "'no', "
                        "'off', 'on', 'true', 'y', 'yes'"
                    ),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.ADD_REMOVE_CANNOT_REMOVE_ITEMS_NOT_IN_THE_CONTAINER,
                    force_code=reports.codes.FORCE,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="cib-bootstrap-options",
                    item_list=["batch-limit", "non-existing"],
                ),
            ]
        )

    def test_validator_errors_forced(self):
        self.config.runner.cib.load(
            crm_config=fixture_crm_config_properties(
                [
                    ("cib-bootstrap-options", {}),
                ]
            )
        )
        self.load_fake_agent_metadata()
        self.config.env.push_cib(
            crm_config=fixture_crm_config_properties(
                [
                    (
                        "cib-bootstrap-options",
                        {
                            "unknown": "property",
                            "enable-acl": "Falsch",
                            "batch-limit": "",
                            "non-existing": "",
                        },
                    ),
                ]
            )
        )

        cluster_property.set_properties(
            self.env_assist.get_env(),
            {
                "unknown": "property",
                "enable-acl": "Falsch",
                "batch-limit": "",
                "non-existing": "",
            },
            [reports.codes.FORCE],
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["unknown"],
                    allowed=ALLOWED_PROPERTIES,
                    option_type="cluster property",
                    allowed_patterns=[],
                ),
                fixture.warn(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="enable-acl",
                    option_value="Falsch",
                    allowed_values=(
                        "a pacemaker boolean value: '0', '1', 'false', 'n', "
                        "'no', "
                        "'off', 'on', 'true', 'y', 'yes'"
                    ),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.warn(
                    reports.codes.ADD_REMOVE_CANNOT_REMOVE_ITEMS_NOT_IN_THE_CONTAINER,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="cib-bootstrap-options",
                    item_list=["batch-limit", "non-existing"],
                ),
            ]
        )

    def test_disable_fencing_warning(self):
        new_properties = {
            "stonith-enabled": "false",
            "fencing-enabled": "false",
        }

        self.config.runner.cib.load(
            crm_config=fixture_crm_config_properties([("id1", {})])
        )
        self.load_fake_agent_metadata()
        self.config.env.push_cib(
            crm_config=fixture_crm_config_properties([("id1", new_properties)])
        )

        cluster_property.set_properties(
            self.env_assist.get_env(),
            new_properties,
            [],
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_name="stonith-enabled",
                    replaced_by=["fencing-enabled"],
                    option_type="cluster property",
                ),
                fixture.warn(
                    reports.codes.NO_STONITH_MEANS_WOULD_BE_LEFT_DUE_TO_PROPERTIES,
                    property_map=new_properties,
                ),
            ]
        )



from typing import (
    Any,
    Optional,
)

from pcs.cli.reports.output import error
from pcs.common import reports
from pcs.common.str_tools import format_optional


def check_is_stonith(
    lib: Any,
    resource_id_list: list[str],
    cmd_to_use: Optional[str] = None,
) -> None:
    if lib.resource.is_any_resource_except_stonith(resource_id_list):
        raise error(
            reports.messages.CommandArgumentTypeMismatch("resources").message
            + format_optional(cmd_to_use, " Please use '{}' instead.")
        )



from pcs_test.tier1.cib_resource.common import ResourceTest
from pcs_test.tools.bin_mock import get_mock_settings

ERRORS_HAVE_OCCURRED = (
    "Error: Errors have occurred, therefore pcs is unable to continue\n"
)


class Enable(ResourceTest):
    def setUp(self):
        super().setUp()
        self.pcs_runner.mock_settings = get_mock_settings()

    def test_enable_disabled_stonith(self):
        self.assert_effect(
            "stonith create S fence_pcsmock_minimal --disabled".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_minimal">
                    <meta_attributes id="S-meta_attributes">
                        <nvpair id="S-meta_attributes-target-role"
                            name="target-role" value="Stopped"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )
        self.assert_effect(
            "stonith enable S".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_minimal">
                    <meta_attributes id="S-meta_attributes"/>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )

    def test_keep_enabled_stonith(self):
        result_xml = """<resources>
            <primitive class="stonith" id="S" type="fence_pcsmock_minimal">
                <operations>
                    <op id="S-monitor-interval-60s" interval="60s"
                        name="monitor"
                    />
                </operations>
            </primitive>
        </resources>"""

        self.assert_effect(
            "stonith create S fence_pcsmock_minimal".split(), result_xml
        )
        self.assert_effect("stonith enable S".split(), result_xml)


class Disable(ResourceTest):
    def setUp(self):
        super().setUp()
        self.pcs_runner.mock_settings = get_mock_settings()

    def test_disable_enabled_stonith(self):
        self.assert_effect(
            "stonith create S fence_pcsmock_minimal".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_minimal">
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )
        self.assert_pcs_fail(
            "stonith disable S".split(),
            (
                "Error: Requested action leaves the cluster with no enabled "
                "means to fence nodes, resulting in the cluster not being able "
                "to recover from certain failure conditions, use --force to "
                "override\n" + ERRORS_HAVE_OCCURRED
            ),
        )

    def test_disable_enabled_stonith_forced(self):
        self.assert_effect(
            "stonith create S fence_pcsmock_minimal".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_minimal">
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
        )
        self.assert_effect(
            "stonith disable S --force".split(),
            """<resources>
                <primitive class="stonith" id="S" type="fence_pcsmock_minimal">
                    <meta_attributes id="S-meta_attributes">
                        <nvpair id="S-meta_attributes-target-role"
                            name="target-role" value="Stopped"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="S-monitor-interval-60s" interval="60s"
                            name="monitor"
                        />
                    </operations>
                </primitive>
            </resources>""",
            stderr_full=(
                "Warning: Requested action leaves the cluster with no enabled "
                "means to fence nodes, resulting in the cluster not being able "
                "to recover from certain failure conditions\n"
            ),
        )

    def test_keep_disabled_stonith(self):
        result_xml = """<resources>
            <primitive class="stonith" id="S" type="fence_pcsmock_minimal">
                <meta_attributes id="S-meta_attributes">
                    <nvpair id="S-meta_attributes-target-role"
                        name="target-role" value="Stopped"
                    />
                </meta_attributes>
                <operations>
                    <op id="S-monitor-interval-60s" interval="60s"
                        name="monitor"
                    />
                </operations>
            </primitive>
        </resources>"""
        self.assert_effect(
            "stonith create S fence_pcsmock_minimal --disabled".split(),
            result_xml,
        )
        self.assert_effect("stonith disable S".split(), result_xml)
